# Copyright (C) 2009-2026 Chris N. Richardson, Garth N. Wells,
# Michal Habera and Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Finite element function spaces and functions."""

from __future__ import annotations

import typing
from collections.abc import Callable, Sequence
from functools import cached_property, singledispatch
from typing import Generic

import numpy as np
import numpy.typing as npt

import basix
import ufl
from dolfinx import cpp as _cpp
from dolfinx import default_scalar_type, jit, la
from dolfinx.fem.dofmap import DofMap
from dolfinx.fem.element import FiniteElement, finiteelement
from dolfinx.geometry import PointOwnershipData
from dolfinx.typing import Real, Scalar

if typing.TYPE_CHECKING:
    from mpi4py import MPI as _MPI

    from dolfinx.mesh import EntityMap as _EntityMap
    from dolfinx.mesh import Mesh


class Constant(ufl.Constant, Generic[Scalar]):
    """A constant with respect to a domain."""

    _cpp_object: (
        _cpp.fem.Constant_complex64
        | _cpp.fem.Constant_complex128
        | _cpp.fem.Constant_float32
        | _cpp.fem.Constant_float64
    )

    def __init__(
        self,
        domain,
        c: float | np.floating | complex | np.complexfloating | Sequence | np.ndarray,
    ):
        """Create a Constant.

        Args:
            domain: DOLFINx or UFL mesh
            c: Value of the constant.
        """
        c = np.asarray(c)
        super().__init__(domain, c.shape)
        try:
            if np.issubdtype(c.dtype, np.complex64):
                self._cpp_object = _cpp.fem.Constant_complex64(c)
            elif np.issubdtype(c.dtype, np.complex128):
                self._cpp_object = _cpp.fem.Constant_complex128(c)
            elif np.issubdtype(c.dtype, np.float32):
                self._cpp_object = _cpp.fem.Constant_float32(c)
            elif np.issubdtype(c.dtype, np.float64):
                self._cpp_object = _cpp.fem.Constant_float64(c)
            else:
                raise RuntimeError("Unsupported dtype")
        except AttributeError:
            raise AttributeError("Constant value must have a dtype attribute.")

    @property
    def value(self):
        """The value of the constant."""
        return self._cpp_object.value

    @value.setter
    def value(self, v: npt.NDArray[Scalar]) -> None:
        np.copyto(self._cpp_object.value, np.asarray(v))

    @property
    def dtype(self) -> npt.DTypeLike:
        """Value dtype of the constant."""
        return np.dtype(self._cpp_object.dtype)

    def __float__(self) -> float:
        """Real representation of the constant."""
        if self.ufl_shape or self.ufl_free_indices:
            raise TypeError("Cannot evaluate a nonscalar expression to a scalar value.")

        return float(self.value)

    def __complex__(self) -> complex:
        """Complex representation of the constant."""
        if self.ufl_shape or self.ufl_free_indices:
            raise TypeError("Cannot evaluate a nonscalar expression to a scalar value.")

        return complex(self.value)


class Expression(Generic[Scalar]):
    """An object for evaluating functions of finite element functions.

    Represents a mathematical expression evaluated at a pre-defined set
    of points on the reference cell. This class closely follows the
    concept of a UFC Expression.

    This functionality can be used to evaluate a gradient of a Function
    at the quadrature points in all cells. This evaluated gradient can
    then be used as input to a non-FEniCS function that calculates a
    material constitutive model.

    """

    _ufl_expression: ufl.core.expr.Expr
    _argument_space: FunctionSpace | None
    _cpp_object: (
        _cpp.fem.Expression_complex64
        | _cpp.fem.Expression_complex128
        | _cpp.fem.Expression_float32
        | _cpp.fem.Expression_float64
    )
    _code: str

    def __init__(
        self,
        e: ufl.core.expr.Expr,
        X: np.ndarray,
        comm: _MPI.Comm | None = None,
        form_compiler_options: dict | None = None,
        jit_options: dict | None = None,
        dtype: npt.DTypeLike | None = None,
        entity_maps: list[_EntityMap] | None = None,
    ):
        """Create an Expression.

        Args:
            e: UFL expression.
            X: Array of points of shape ``(num_points, tdim)`` on the
                reference element.
            comm: Communicator that the Expression is defined on.
            form_compiler_options: Options used in FFCx compilation of
                this Expression. Run ``ffcx --help`` in the commandline
                to see all available options.
            jit_options: Options controlling JIT compilation of C code.
            dtype: Type of the Expression values. If ``None``,
                the dtype is deduced from the UFL expression.
            entity_maps: Maps between different meshes.

        Note:
            This wrapper is responsible for the FFCx compilation of the
            UFL Expr and attaching the correct data to the underlying
            C++ Expression.
        """
        assert X.ndim < 3
        num_points = X.shape[0] if X.ndim == 2 else 1
        _X = np.reshape(X, (num_points, -1))

        # Get MPI communicator
        if comm is None:
            try:
                domains = ufl.domain.extract_domains(e)
                if len(domains) == 0:
                    raise RuntimeError("Could not extract MPI communicator for Expression.")
                domain = domains[0]
                assert isinstance(domain, ufl.Mesh)
                mesh = domain.ufl_cargo()
                comm = mesh.comm
            except AttributeError:
                print(
                    "Could not extract MPI communicator for Expression. "
                    + "Maybe you need to pass a communicator?"
                )
                raise

        # Attempt to deduce dtype
        if dtype is None:
            dtype = getattr(e, "dtype", default_scalar_type)

        # Compile UFL expression with JIT
        if form_compiler_options is None:
            form_compiler_options = dict()
        form_compiler_options["scalar_type"] = dtype
        self._ufcx_expression, module, self._code = jit.ffcx_jit(
            comm, (e, _X), form_compiler_options=form_compiler_options, jit_options=jit_options
        )
        self._ufl_expression = e

        # Prepare coefficients data. For every coefficient in expression
        # take its C++ object.
        original_coefficients = ufl.algorithms.extract_coefficients(e)
        coeffs = [
            original_coefficients[
                self._ufcx_expression.original_coefficient_positions[i]
            ]._cpp_object
            for i in range(self._ufcx_expression.num_coefficients)
        ]
        ufl_constants = ufl.algorithms.analysis.extract_constants(e)
        constants = [constant._cpp_object for constant in ufl_constants]
        arguments = ufl.algorithms.extract_arguments(e)
        if len(arguments) == 0:
            self._argument_space = None
        elif len(arguments) == 1:
            self._argument_space = arguments[0].ufl_function_space()._cpp_object
        else:
            raise RuntimeError("Expressions with more that one Argument not allowed.")

        def _create_expression(dtype):
            if np.issubdtype(dtype, np.float32):
                return _cpp.fem.create_expression_float32
            elif np.issubdtype(dtype, np.float64):
                return _cpp.fem.create_expression_float64
            elif np.issubdtype(dtype, np.complex64):
                return _cpp.fem.create_expression_complex64
            elif np.issubdtype(dtype, np.complex128):
                return _cpp.fem.create_expression_complex128
            else:
                raise NotImplementedError(f"Type {dtype} not supported.")

        _entity_maps = (
            [entity_map._cpp_object for entity_map in entity_maps]
            if entity_maps is not None
            else []
        )
        ffi = module.ffi
        self._cpp_object = _create_expression(dtype)(
            ffi.cast("uintptr_t", ffi.addressof(self._ufcx_expression)),
            coeffs,
            constants,
            _entity_maps,
            self.argument_space,
        )

    def eval(
        self,
        mesh: Mesh,
        entities: npt.NDArray[np.int32],
        values: npt.NDArray[Scalar] | None = None,
    ) -> npt.NDArray[Scalar]:
        """Evaluate Expression on entities.

        Args:
            mesh: Mesh to evaluate Expression on.
            entities: Entities to evaluate the Expression over. For
                cells, it is a list of cell indices. For facets, it is a
                2D array of (cell index, local facet index).
            values: Array to fill with evaluated values. If ``None``,
                storage will be allocated. Otherwise it must have shape
                ``(entities.shape[0], num_points, *value_shape)`` if
                there is not argument function and shape
                ``(entities.shape[0], num_points, *value_shape,
                argument_space_dim)`` if the Expression does have an
                argument function.

        Returns:
            Expression evaluated at points for ``entities``. Shape is
            ``(entities.shape[0], num_points, *value_shape)`` if there
            is no argument function, or shape is ``(entities.shape[0],
            num_points, *value_shape, argument_space_dim)`` if the
            Expression does have an argument function.
        """
        _entities = np.asarray(entities, dtype=np.int32)
        if (tdim := mesh.topology.dim) != (expr_dim := self._cpp_object.X().shape[1]):
            assert expr_dim == tdim - 1
            assert entities.ndim == 2, (
                "entities list should have two dimensions for expression evaluation on facets."
            )

        if self.argument_space is None:
            values_shape = (_entities.shape[0], self.X().shape[0], *self.value_shape)
        else:
            values_shape = (
                _entities.shape[0],
                self.X().shape[0],
                *self.value_shape,
                self.argument_space.element.space_dimension,
            )

        # Allocate memory for result if u was not provided
        if values is None:
            values = np.zeros(values_shape, dtype=self.dtype)
        else:
            if values.shape != values_shape:
                raise TypeError("Passed values array does not have correct shape.")
            if values.dtype != self.dtype:
                raise TypeError("Passed values array does not have correct dtype.")

        constants = _cpp.fem.pack_constants(self._cpp_object)
        coeffs = _cpp.fem.pack_coefficients(self._cpp_object, mesh._cpp_object, _entities)
        _cpp.fem.tabulate_expression(
            values, self._cpp_object, constants, coeffs, mesh._cpp_object, _entities
        )
        return values

    def X(self) -> npt.NDArray:
        """Evaluation points on the reference cell."""
        return self._cpp_object.X()

    @property
    def ufl_expression(self) -> ufl.core.expr.Expr:
        """Original UFL Expression."""
        return self._ufl_expression

    @property
    def value_shape(self) -> np.ndarray:
        """Value shape of the Expression."""
        return self._cpp_object.value_shape

    @property
    def value_size(self) -> int:
        """Value size of the expression."""
        return self._cpp_object.value_size

    @property
    def argument_space(self) -> FunctionSpace | None:
        """Argument function space if Expression has argument."""
        return self._argument_space

    @property
    def ufcx_expression(self):
        """The compiled ufcx_expression object."""
        return self._ufcx_expression

    @property
    def code(self) -> str:
        """C code strings."""
        return self._code

    @property
    def dtype(self) -> npt.DTypeLike:
        """Expression value dtype."""
        return np.dtype(self._cpp_object.dtype)


class Function(ufl.Coefficient, Generic[Scalar]):
    """A finite element function.

    A finite element function is represented by a function space
    (domain, element and dofmap) and a vector holding the
    degrees-of-freedom.

    """

    _cpp_object: (
        _cpp.fem.Function_complex64
        | _cpp.fem.Function_complex128
        | _cpp.fem.Function_float32
        | _cpp.fem.Function_float64
    )

    _x: la.Vector[Scalar]

    def __init__(
        self,
        V: FunctionSpace,
        x: la.Vector[Scalar] | None = None,
        name: str | None = None,
        dtype: npt.DTypeLike | None = None,
    ):
        """Initialize a finite element Function.

        Args:
            V: The function space that the Function is defined on.
            x: Function degree-of-freedom vector. Typically required
                only when reading a saved Function from file.
            name: Function name.
            dtype: Scalar type. Is not set, the DOLFINx default scalar
                type is used.
        """
        if x is not None:
            if dtype is None:
                dtype = x.array.dtype
            else:
                assert x.array.dtype == dtype, "Incompatible Vector and dtype."
        else:
            if dtype is None:
                dtype = default_scalar_type

        assert np.issubdtype(V.element.dtype, np.dtype(dtype).type(0).real.dtype), (
            "Incompatible FunctionSpace dtype and requested dtype."
        )

        # Create cpp Function
        def functiontype(dtype):
            if np.issubdtype(dtype, np.float32):
                return _cpp.fem.Function_float32
            elif np.issubdtype(dtype, np.float64):
                return _cpp.fem.Function_float64
            elif np.issubdtype(dtype, np.complex64):
                return _cpp.fem.Function_complex64
            elif np.issubdtype(dtype, np.complex128):
                return _cpp.fem.Function_complex128
            else:
                raise NotImplementedError(f"Type {dtype} not supported.")

        if x is not None:
            self._cpp_object = functiontype(dtype)(V._cpp_object, x._cpp_object)  # type: ignore
        else:
            self._cpp_object = functiontype(dtype)(V._cpp_object)  # type: ignore

        # Initialize the ufl.FunctionSpace
        super().__init__(V.ufl_function_space())

        # Set name
        if name is None:
            self.name = "f"
        else:
            self.name = name

        # Store DOLFINx FunctionSpace object
        self._V = V

        # Store Python wrapper around the underlying Vector
        self._x = la.Vector(self._cpp_object.x)

    @property
    def function_space(self) -> FunctionSpace:
        """FunctionSpace that the Function is defined on."""
        return self._V

    def eval(
        self,
        x: npt.ArrayLike,
        cells: npt.NDArray[np.int32],
        u: None | npt.NDArray[Scalar] = None,
        tol: float = 1.0e-6,
        maxit: int = 15,
    ) -> npt.NDArray[Scalar]:
        """Evaluate Function at points x.

        Args:
            x: Points with shape (num_points, 3)
            cells: Array with cell indices, with shape (num_points,), where
              cell[i] is the index of the cell containing point x[i].
              If the cell index is negative the point is ignored.
            u: Array to put evaluated data in.
            tol: Tolerance for convergence in Newton method for
                nonaffine pullbacks.
            maxit: Maximum number of Newton iterations for
                nonaffine pullbacks.
        """
        # Make sure input coordinates are a NumPy array
        _x = np.asarray(x, dtype=self._V.mesh.geometry.x.dtype)
        assert _x.ndim < 3
        if len(_x) == 0:
            _x = np.zeros((0, 3), dtype=self._V.mesh.geometry.x.dtype)
        else:
            shape0 = _x.shape[0] if _x.ndim == 2 else 1
            _x = np.reshape(_x, (shape0, -1))
        num_points = _x.shape[0]
        if _x.shape[1] != 3:
            raise ValueError("Coordinate(s) for Function evaluation must have length 3.")

        # Make sure cells are a NumPy array
        _cells = np.asarray(cells, dtype=np.int32)
        assert _cells.ndim < 2
        num_points_c = _cells.shape[0] if _cells.ndim == 1 else 1
        _cells = np.reshape(_cells, num_points_c)

        # Allocate memory for return value if not provided
        if u is None:
            value_size = self._V.value_size
            u = np.empty((num_points, value_size), self.dtype)

        self._cpp_object.eval(_x, _cells, u, tol, maxit)  # type: ignore
        if num_points == 1:
            u = np.reshape(u, (-1,))
        return u

    def interpolate_nonmatching(
        self,
        u0: Function[Scalar],
        cells: npt.NDArray[np.int32],
        interpolation_data: PointOwnershipData,
        tol: float = 1e-6,
        maxit: int = 15,
    ) -> None:
        """Interpolate a Function on a non-matching mesh.

        Args:
            u0: Function to interpolate.
            cells: The cells to interpolate over. If ``None`` then all
                cells are interpolated over.
            interpolation_data: Data needed to interpolate functions
                defined on other meshes. Created by
                :func:`dolfinx.fem.create_interpolation_data`.
            tol: Tolerance for convergence in Newton method for nonaffine
                pullbacks. Ignored if mesh geometry is affine.
            maxit: Maximum number of iterations for nonaffine pullback.
                Ignored if mesh geometry is affine.
        """
        self._cpp_object.interpolate(
            u0._cpp_object, cells, tol, maxit, interpolation_data._cpp_object
        )  # type: ignore

    def interpolate(
        self,
        u0: Callable | Expression[Scalar] | Function[Scalar],
        cells0: npt.NDArray[np.int32] | None = None,
        cells1: npt.NDArray[np.int32] | None = None,
    ) -> None:
        """Interpolate an expression.

        Args:
            u0: Callable function, Expression or Function to
               interpolate.
            cells0: Cells in mesh associated with ``u0`` to interpolate
                over. If ``None`` then all cells are interpolated over.
            cells1: Cells in the mesh associated with ``self`` to
                interpolate over. If ``None``, then taken to be the same
                cells as ``cells0``. If ``cells1`` is not ``None`` it
                must have the same length as ``cells0``.
        """

        @singledispatch
        def _interpolate(u0):
            """Interpolate a cpp.fem.Function."""
            self._cpp_object.interpolate(u0, cells0, cells1)

        @_interpolate.register(Function)
        def _(u0: Function):
            """Interpolate a fem.Function."""
            self._cpp_object.interpolate(u0._cpp_object, cells0, cells1)

        @_interpolate.register(int)
        def _(u0_ptr: int):
            """Interpolate using a pointer to a function f(x)."""
            self._cpp_object.interpolate_ptr(u0_ptr, cells0)

        @_interpolate.register(Expression)
        def _(e0: Expression):
            """Interpolate a fem.Expression."""
            self._cpp_object.interpolate_expr(e0._cpp_object, cells0, cells1)

        try:
            # u is a Function or Expression (or pointer to one)
            _interpolate(u0)
        except TypeError:
            # u0 is callable
            assert callable(u0)
            x = _cpp.fem.interpolation_coords(
                self._V.element._cpp_object, self._V.mesh.geometry._cpp_object, cells0
            )
            self._cpp_object.interpolate_f(np.asarray(u0(x), dtype=self.dtype), cells0)

    def copy(self) -> Function[Scalar]:
        """Create a copy of the Function.

        The function space is shared and the degree-of-freedom vector is
        copied.

        Returns:
            A new Function with a copy of the degree-of-freedom vector.
        """
        return Function(
            self.function_space,
            la.Vector(type(self.x._cpp_object)(self.x._cpp_object)),
            name=self.name,
        )

    @property
    def x(self) -> la.Vector[Scalar]:
        """Vector holding the degrees-of-freedom."""
        return self._x

    @property
    def dtype(self) -> npt.DTypeLike:
        """Function value dtype."""
        return np.dtype(self._cpp_object.x.array.dtype)

    @property
    def name(self) -> str:
        """Name of the Function."""
        return self._cpp_object.name  # type: ignore

    @name.setter
    def name(self, name):
        self._cpp_object.name = name

    def __str__(self) -> str:
        """Pretty print representation."""
        return self.name

    def sub(self, i: int) -> Function[Scalar]:
        """Return a sub-function (a view into the ``Function``).

        Sub-functions are indexed ``i = 0, ..., N-1``, where ``N`` is
        the number of sub-spaces.

        Args:
            i: Index of the sub-function to extract.

        Returns:
            A view into the parent ``Function``.

        Note:
            If the sub-Function is re-used, for performance reasons the
            returned ``Function`` should be stored by the caller to
            avoid repeated re-computation of the subspace.
        """
        return Function(self._V.sub(i), self.x, name=f"{self!s}_{i}")

    def split(self) -> tuple[Function[Scalar], ...]:
        """Extract (any) sub-functions.

        A sub-function can be extracted from a discrete function that is
        in a mixed, vector, or tensor FunctionSpace. The sub-function
        resides in the subspace of the mixed space.

        Returns:
            First level of subspaces of the function space.
        """
        num_sub_spaces = self.function_space.num_sub_spaces
        if num_sub_spaces == 1:
            raise RuntimeError("No subfunctions to extract")
        return tuple(self.sub(i) for i in range(num_sub_spaces))

    def collapse(self) -> Function[Scalar]:
        """Create a collapsed version of this Function."""
        u_collapsed = self._cpp_object.collapse()  # type: ignore
        V_collapsed = FunctionSpace(
            self.function_space._mesh,
            self.ufl_element(),  # type: ignore
            u_collapsed.function_space,
        )
        return Function(V_collapsed, la.Vector(u_collapsed.x))


class ElementMetaData(typing.NamedTuple):
    """Data for representing a finite element.

    :param family: Element type.
    :param degree: Polynomial degree of the element.
    :param shape: Shape for vector/tensor valued elements that are
        constructed from blocked scalar elements (e.g., Lagrange).
    :param symmetry: Symmetry option for blocked tensor elements.
    """

    family: str
    degree: int
    shape: tuple[int, ...] | None = None
    symmetry: bool | None = None


def functionspace(
    mesh: Mesh,
    element: (
        ufl.finiteelement.AbstractFiniteElement
        | ElementMetaData
        | tuple[str, int]
        | tuple[str, int, tuple]
        | tuple[str, int, tuple, bool]
    ),
) -> FunctionSpace:
    """Create a finite element function space.

    Args:
        mesh: Mesh that space is defined on.
        element: Finite element description.

    Returns:
        A function space.
    """
    # Create UFL element
    dtype = mesh.geometry.x.dtype
    try:
        e = ElementMetaData(*element)  # type: ignore
        ufl_e = basix.ufl.element(
            e.family,
            mesh.basix_cell(),  # type: ignore
            e.degree,
            shape=e.shape,
            symmetry=e.symmetry,
            dtype=dtype,
        )
    except TypeError:
        ufl_e = element  # type: ignore

    # Check that element and mesh cell types match
    if ((domain := mesh.ufl_domain()) is None) or ufl_e.cell != domain.ufl_cell():
        raise ValueError("Non-matching UFL cell and mesh cell shapes.")
    # Create DOLFINx objects
    element = finiteelement(mesh.topology.cell_type, ufl_e, dtype)  # type: ignore

    if ufl_e.is_real:
        cpp_dofmap = _cpp.fem.build_real_element_dofmap(
            mesh.topology._cpp_object,
            element.basix_element.entity_dofs,  # type: ignore
            element.basix_element.entity_closure_dofs,  # type: ignore
            int(np.prod(element.value_shape)),  # type: ignore
        )
    else:
        cpp_dofmap = _cpp.fem.create_dofmap(
            mesh.comm,
            mesh.topology._cpp_object,
            element._cpp_object,  # type: ignore
        )
    assert np.issubdtype(mesh.geometry.x.dtype, element.dtype), (  # type: ignore
        "Mesh and element dtype are not compatible."
    )

    # Initialize the cpp.FunctionSpace
    try:
        cppV = _cpp.fem.FunctionSpace_float64(mesh._cpp_object, element._cpp_object, cpp_dofmap)  # type: ignore
    except TypeError:
        cppV = _cpp.fem.FunctionSpace_float32(mesh._cpp_object, element._cpp_object, cpp_dofmap)  # type: ignore

    return FunctionSpace(mesh, ufl_e, cppV)


class FunctionSpace(ufl.FunctionSpace, Generic[Real]):
    """A space on which Functions (fields) can be defined."""

    _cpp_object: _cpp.fem.FunctionSpace_float32 | _cpp.fem.FunctionSpace_float64
    _mesh: Mesh[Real]

    def __init__(
        self,
        mesh: Mesh[Real],
        element: ufl.finiteelement.AbstractFiniteElement,
        cppV: (_cpp.fem.FunctionSpace_float32 | _cpp.fem.FunctionSpace_float64),
    ):
        """Create a finite element function space.

        Note:
            This initialiser is for internal use and not normally called
            in user code. Use :func:`functionspace` to create a function
            space.

        Args:
            mesh: Mesh that space is defined on.
            element: UFL finite element.
            cppV: Compiled C++ function space.
        """
        if mesh._cpp_object is not cppV.mesh:
            raise RuntimeError("Meshes do not match in function space initialisation.")
        ufl_domain = mesh.ufl_domain()
        self._cpp_object = cppV
        self._mesh = mesh
        super().__init__(ufl_domain, element)

    def clone(self) -> FunctionSpace[Real]:
        """Create a FunctionSpace which shares data with this space.

        The new space has a different unique integer ID.

        Create a new FunctionSpace :math:`W` which shares data with this
        FunctionSpace :math:`V`, but with a different unique integer ID.

        This function is helpful for defining mixed problems and using
        blocked linear algebra. For example, a matrix block defined on
        the spaces :math:`V \\times W` where, :math:`V` and :math:`W`
        are defined on the same finite element and mesh can be
        identified as an off-diagonal block whereas the :math:`V \\times
        V` and :math:`V \\times V` matrices can be identified as
        diagonal blocks. This is relevant for the handling of boundary
        conditions.

        Returns:
            A new function space that shares data
        """  # noqa: D301
        try:
            Vcpp = _cpp.fem.FunctionSpace_float64(
                self._cpp_object.mesh, self._cpp_object.element, self._cpp_object.dofmap
            )  # type: ignore
        except TypeError:
            Vcpp = _cpp.fem.FunctionSpace_float32(
                self._cpp_object.mesh, self._cpp_object.element, self._cpp_object.dofmap
            )  # type: ignore
        return FunctionSpace(self._mesh, self.ufl_element(), Vcpp)

    @property
    def num_sub_spaces(self) -> int:
        """Number of sub spaces."""
        return self.element.num_sub_elements

    def sub(self, i: int) -> FunctionSpace[Real]:
        """Return the i-th sub space.

        Args:
            i: Index of the subspace to extract.

        Returns:
            A subspace.

        Note:
            If the subspace is re-used, for performance reasons the
            returned subspace should be stored by the caller to avoid
            repeated re-computation of the subspace.
        """
        assert self.ufl_element().num_sub_elements > i
        sub_element = self.ufl_element().sub_elements[i]
        cppV_sub = self._cpp_object.sub([i])  # type: ignore
        return FunctionSpace(self._mesh, sub_element, cppV_sub)

    def component(self):
        """Return the component relative to the parent space."""
        return self._cpp_object.component()  # type: ignore

    def contains(self, V) -> bool:
        """Check if a space is contained in, or is the same as, this space.

        Args:
            V: The space to check to for inclusion.

        Returns:
           `` True`` if ``V`` is contained in, or is the same as, this
           space.
        """
        return self._cpp_object.contains(V._cpp_object)  # type: ignore

    def __eq__(self, other):
        """Comparison for equality."""
        return super().__eq__(other) and self._cpp_object == other._cpp_object

    def __ne__(self, other):
        """Comparison for inequality."""
        return super().__ne__(other) or self._cpp_object != other._cpp_object

    def ufl_function_space(self) -> ufl.FunctionSpace:
        """UFL function space."""
        return self

    @cached_property
    def element(self) -> FiniteElement[Real]:
        """Function space finite element."""
        return FiniteElement(self._cpp_object.element)

    @property
    def dofmap(self) -> DofMap:
        """Degree-of-freedom map associated with the function space."""
        return DofMap(self._cpp_object.dofmap)

    @property
    def dofmaps(self) -> list[DofMap]:
        """The geometry dofmaps, one per cell type."""
        return [DofMap(_o) for _o in self._cpp_object.dofmaps]

    @property
    def mesh(self) -> Mesh[Real]:
        """Mesh on which the function space is defined."""
        return self._mesh

    def collapse(self) -> tuple[FunctionSpace[Real], list[npt.NDArray[np.int32]]]:
        """Create a new function space by collapsing a subspace.

        Returns:
            A new function space and the map from new to old
            degrees-of-freedom.
        """
        cpp_space, dofs = self._cpp_object.collapse()
        V = FunctionSpace(self._mesh, self.ufl_element(), cpp_space)
        return V, dofs

    def tabulate_dof_coordinates(self) -> npt.NDArray[Real]:
        """Tabulate coordinates of function space degrees-of-freedom.

        Returns:
            Coordinates of the degrees-of-freedom.

        Note:
            This method is only for elements with point evaluation
            degrees-of-freedom.
        """
        return self._cpp_object.tabulate_dof_coordinates()



"""Poisson problem."""

from basix.ufl import element
from ufl import (
    Coefficient,
    Constant,
    FunctionSpace,
    Mesh,
    TestFunction,
    TrialFunction,
    dx,
    grad,
    inner,
)

e = element("Lagrange", "tetrahedron", 2)
coord_element = element("Lagrange", "tetrahedron", 1, shape=(3,))
mesh = Mesh(coord_element)

V = FunctionSpace(mesh, e)

u = TrialFunction(V)
v = TestFunction(V)
f = Coefficient(V)
kappa = Constant(mesh)

a = kappa * inner(grad(u), grad(v)) * dx



# Copyright (C) 2019 Matthew Scroggs
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Test that the vectors in vector spaces are correctly oriented."""

from mpi4py import MPI

import numpy as np
import pytest

import ufl
from basix.ufl import element
from dolfinx import default_real_type
from dolfinx.fem import Function, functionspace
from dolfinx.mesh import create_mesh


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("space_type", ["RT"])
@pytest.mark.parametrize("order", [1, 2, 3, 4, 5])
def test_div_conforming_triangle(space_type, order):
    """Checks that the vectors in div conforming spaces on a triangle are correctly oriented."""

    # Create simple triangle mesh
    def perform_test(points, cells):
        domain = ufl.Mesh(element("Lagrange", "triangle", 1, shape=(2,), dtype=default_real_type))
        mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
        V = functionspace(mesh, (space_type, order))
        f = Function(V)
        x = f.x.array
        output = []
        for dof in range(len(x)):
            x[:] = 0.0
            x[dof] = 1
            points = np.array([[0.5, 0.5, 0], [0.5, 0.5, 0]])
            cells = np.array([0, 1])
            result = f.eval(points, cells)
            normal = np.array([-1.0, 1.0])
            output.append(result.dot(normal))
        return output

    points = np.array([[0, 0], [1, 0], [1, 1], [0, 1]], dtype=default_real_type)
    cells = np.array([[0, 1, 2], [2, 3, 0]])
    result = perform_test(points, cells)
    for i, j in result:
        assert i == pytest.approx(j, abs=1.0e-5)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("space_type", ["RT"])
@pytest.mark.parametrize("order", [1, 2, 3, 4, 5])
def test_div_conforming_tetrahedron(space_type, order):
    """Checks that the vectors in div conforming spaces on a tetrahedron are correctly oriented."""

    # Create simple tetrahedron cell mesh
    def perform_test(points, cells):
        domain = ufl.Mesh(
            element("Lagrange", "tetrahedron", 1, shape=(3,), dtype=default_real_type)
        )
        mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
        V = functionspace(mesh, (space_type, order))
        f = Function(V)
        output = []
        x = f.x.array
        for dof in range(len(x)):
            x[:] = 0.0
            x[dof] = 1
            points = np.array([[1 / 3, 1 / 3, 1 / 3], [1 / 3, 1 / 3, 1 / 3]])
            cells = np.array([0, 1])
            result = f.eval(points, cells)
            normal = np.array([1.0, 1.0, 1.0])
            output.append(result.dot(normal))
        return output

    points = np.array(
        [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 1]], dtype=default_real_type
    )
    cells = np.array([[0, 1, 2, 3], [1, 3, 2, 4]])
    result = perform_test(points, cells)
    for i, j in result:
        assert i == pytest.approx(j, rel=1.0e-6, abs=1.0e-4)



# Copyright (C) 2021 Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Unit tests for the function library."""

from mpi4py import MPI

import numpy as np
import pytest

import ufl
from dolfinx import default_scalar_type
from dolfinx.fem import Constant, assemble_scalar, form
from dolfinx.mesh import create_unit_cube, create_unit_interval, create_unit_square


def test_facet_area1D():
    mesh = create_unit_interval(MPI.COMM_WORLD, 10)

    # NOTE: Area of a vertex is defined to 1 in ufl
    c0 = ufl.FacetArea(mesh)
    c = Constant(mesh, default_scalar_type(1))

    ds = ufl.Measure("ds", domain=mesh)
    a0 = mesh.comm.allreduce(assemble_scalar(form(c * ds)), op=MPI.SUM)
    a = mesh.comm.allreduce(assemble_scalar(form(c0 * ds)), op=MPI.SUM)
    assert np.isclose(a.real, 2)
    assert np.isclose(a0.real, 2)


@pytest.mark.parametrize(
    "mesh_factory",
    [
        (create_unit_square, (MPI.COMM_WORLD, 3, 3), 1.0 / 3),
        #   (create_unit_square,
        #   (MPI.COMM_WORLD, 3, 3, CellType.quadrilateral), 1. / 3),
        (create_unit_cube, (MPI.COMM_WORLD, 3, 3, 3), 1 / 18.0),
        #   (create_unit_cube,
        #   (MPI.COMM_WORLD, 3, 3, 3, CellType.hexahedron), 1. / 9)
    ],
)
def test_facet_area(mesh_factory):
    """Compute facet area of cell. UFL currently only supports affine
    cells for this computation.
    """
    # NOTE: UFL only supports facet area calculations of affine cells
    func, args, exact_area = mesh_factory
    mesh = func(*args)
    c0 = ufl.FacetArea(mesh)
    c = Constant(mesh, default_scalar_type(1))
    tdim = mesh.topology.dim
    num_faces = 4 if tdim == 2 else 6

    ds = ufl.Measure("ds", domain=mesh)
    a = mesh.comm.allreduce(assemble_scalar(form(c * ds)), op=MPI.SUM)
    a0 = mesh.comm.allreduce(assemble_scalar(form(c0 * ds)), op=MPI.SUM)
    assert np.isclose(a.real, num_faces)
    assert np.isclose(a0.real, num_faces * exact_area)



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.6
# ---

# # Stokes equations using Taylor-Hood elements
#
# ```{admonition} Download sources
# :class: download
# * {download}`Python script <./demo_stokes.py>`
# * {download}`Jupyter notebook <./demo_stokes.ipynb>`
# ```
# It shows how to solve the Stokes problem using Taylor-Hood elements
# using different linear solvers:
#
# 1. [Block preconditioner using PETSc Nest data structures using
#    {py:class}`LinearProblem <dolfinx.fem.petsc.LinearProblem>`
#    ](#high-level-nested-matrix-solver)
# 1. [Block preconditioner using PETSc Nest data structures using
#    PETSc directly](#low-level-nested-matrix-solver)
# 1. [Block preconditioner with the `u` and `p` fields stored block-wise
#    in a single matrix](#monolithic-block-iterative-solver)
# 1. [Direct solver with the `u` and `p` fields stored block-wise in a
#    single matrix](#monolithic-block-direct-solver)
# 1. [Direct solver with the `u` and `p` fields stored in a mixed fashion a
#    single matrix](#non-blocked-direct-solver)
#
# ## Equation and problem definition
#
# ### Strong formulation
#
# $$
# \begin{aligned}
#   - \nabla \cdot (\nabla u + p I) &= f \quad {\rm in} \ \Omega,\\
#   \nabla \cdot u &= 0 \quad {\rm in} \ \Omega.
# \end{aligned}
# $$
#
# with conditions on the boundary $\partial \Omega = \Gamma_{D} \cup
# \Gamma_{N}$ of the form:
#
# $$
# \begin{aligned}
#   u &= u_0 \quad {\rm on} \ \Gamma_{D},\\
#   \nabla u \cdot n + p n &= g \,   \quad\;\; {\rm on} \ \Gamma_{N}.
# \end{aligned}
# $$
#
# ```{note}
# The sign of the pressure has been changed from the usual
# definition. This is to generate have a symmetric system
# of equations.
# ```
#
# ### Weak formulation
#
# The weak formulation reads: find $(u, p) \in V \times Q$ such that
#
# $$
# a((u, p), (v, q)) = L((v, q)) \quad \forall  (v, q) \in V \times Q
# $$
#
# where
#
# $$
# \begin{aligned}
#   a((u, p), (v, q)) &:= \int_{\Omega} \nabla u \cdot \nabla v -
#            \nabla \cdot v \ p + \nabla \cdot u \ q \, {\rm d} x,\\
#   L((v, q)) &:= \int_{\Omega} f \cdot v \, {\rm d} x + \int_{\partial
#            \Omega_N} g \cdot v \, {\rm d} s.
# \end{aligned}
# $$
#
# ### Domain and boundary conditions
#
# We consider the lid-driven cavity problem with the following
# domain and boundary conditions:
#
# - $\Omega := [0,1]\times[0,1]$ (a unit square)
# - $\Gamma_D := \partial \Omega$
# - $u_0 := (1, 0)^\top$ at $x_1 = 1$ and $u_0 = (0, 0)^\top$ otherwise
# - $f := (0, 0)^\top$
#
#
# ## Implementation
#
# The required modules are first imported:

# +
from mpi4py import MPI
from petsc4py import PETSc

import numpy as np

import ufl
from basix.ufl import element, mixed_element
from dolfinx import default_real_type, la
from dolfinx.fem import (
    Constant,
    Function,
    bcs_by_block,
    dirichletbc,
    extract_function_spaces,
    form,
    functionspace,
    locate_dofs_topological,
)
from dolfinx.fem.petsc import (
    LinearProblem,
    apply_lifting,
    assemble_matrix,
    assemble_vector,
    create_vector,
    set_bc,
)
from dolfinx.io import XDMFFile
from dolfinx.la.petsc import create_vector_wrap
from dolfinx.mesh import CellType, create_rectangle, locate_entities_boundary

# -

# We create a {py:class}`Mesh <dolfinx.mesh.Mesh>`, define functions for
# locating geometrically subsets of the boundary, and define a function
# for the  velocity on the lid:

# +
# Create mesh
msh = create_rectangle(
    MPI.COMM_WORLD, [np.array([0, 0]), np.array([1, 1])], (32, 32), CellType.triangle
)


# Function to mark x = 0, x = 1 and y = 0
def noslip_boundary(x):
    """Mark no-slip boundaries where x = 0, x = 1, and y = 0."""
    return np.isclose(x[0], 0.0) | np.isclose(x[0], 1.0) | np.isclose(x[1], 0.0)


# Function to mark the lid (y = 1)
def lid(x):
    """Mark lid boundary where y = 1."""
    return np.isclose(x[1], 1.0)


# Lid velocity
def lid_velocity_expression(x):
    """Expression for lid velocity."""
    return np.stack((np.ones(x.shape[1]), np.zeros(x.shape[1])))


# -

# Two {py:class}`function spaces <dolfinx.fem.FunctionSpace>` are
# defined using different finite elements. `P2` corresponds to a
# continuous piecewise quadratic basis (vector) and `P1` to a continuous
# piecewise linear basis (scalar).

gdim = msh.geometry.dim
P2 = element("Lagrange", msh.basix_cell(), degree=2, shape=(gdim,), dtype=default_real_type)
P1 = element("Lagrange", msh.basix_cell(), degree=1, dtype=default_real_type)
V, Q = functionspace(msh, P2), functionspace(msh, P1)

# Boundary conditions for the velocity field are defined:

# +
# No-slip condition on boundaries where x = 0, x = 1, and y = 0
noslip = np.zeros(gdim, dtype=PETSc.ScalarType)  # type: ignore
facets = locate_entities_boundary(msh, 1, noslip_boundary)
bc0 = dirichletbc(noslip, locate_dofs_topological(V, 1, facets), V)

# Driving (lid) velocity condition on top boundary (y = 1)
lid_velocity = Function(V)
lid_velocity.interpolate(lid_velocity_expression)
facets = locate_entities_boundary(msh, 1, lid)
bc1 = dirichletbc(lid_velocity, locate_dofs_topological(V, 1, facets))

# Collect Dirichlet boundary conditions
bcs = [bc0, bc1]
# -

# The bilinear and linear forms for the Stokes equations are defined
# using a blocked structure:

# +
# Define variational problem
(u, p) = ufl.TrialFunction(V), ufl.TrialFunction(Q)
(v, q) = ufl.TestFunction(V), ufl.TestFunction(Q)
f = Constant(msh, (PETSc.ScalarType(0), PETSc.ScalarType(0)))  # type: ignore

a_ufl = [
    [ufl.inner(ufl.grad(u), ufl.grad(v)) * ufl.dx, ufl.inner(p, ufl.div(v)) * ufl.dx],
    [ufl.inner(ufl.div(u), q) * ufl.dx, None],
]
a = form(a_ufl)
L_ufl = [ufl.inner(f, v) * ufl.dx, ufl.ZeroBaseForm((q,))]
L = form(L_ufl)
# -

# A block-diagonal preconditioner will be used with the iterative
# solvers for this problem:

a_p11 = form(ufl.inner(p, q) * ufl.dx)
a_p = [[a[0][0], None], [None, a_p11]]


# ### High-level nested matrix solver
#
# We first use the high-level {py:class}`LinearProblem
# <dolfinx.fem.petsc.LinearProblem>` class which uses PETSc to solve
# the linear problem. Details on the preconditioner setup are given in
# {py:func}`nested_iterative_solver_low_level` below.


def nested_iterative_solver_high_level():
    """Solve Stokes problem using nest matrices and an iterative solver.

    Uses high-level DOLFINx functionality.
    """
    problem = LinearProblem(
        a_ufl,
        L_ufl,
        kind="nest",
        bcs=bcs,
        P=a_p,
        petsc_options_prefix="demo_stokes__nested_iterative_solver_high_level_",
        petsc_options={
            "ksp_type": "minres",
            "ksp_rtol": 1e-9,
            "pc_type": "fieldsplit",
            "pc_fieldsplit_type": "additive",
            "fieldsplit_0_ksp_type": "preonly",
            "fieldsplit_0_pc_type": "gamg",
            "fieldsplit_1_ksp_type": "preonly",
            "fieldsplit_1_pc_type": "jacobi",
        },
    )

    # The pressure field is determined only up to a constant. We supply
    # a vector that spans the nullspace to the solver, and any component
    # of the solution in this direction will be eliminated during the
    # solution process.
    null_vec = create_vector(extract_function_spaces(L), "nest")

    # Set velocity part to zero and the pressure part to a non-zero
    # constant
    null_vecs = null_vec.getNestSubVecs()
    null_vecs[0].set(0.0), null_vecs[1].set(1.0)

    # Normalize the vector that spans the nullspace, create a nullspace
    # object, and attach it to the matrix
    null_vec.normalize()
    nsp = PETSc.NullSpace().create(vectors=[null_vec])
    problem.A.setNullSpace(nsp)

    A00 = problem.A.getNestSubMatrix(0, 0)
    A00.setOption(PETSc.Mat.Option.SPD, True)

    P00, P11 = problem.P_mat.getNestSubMatrix(0, 0), problem.P_mat.getNestSubMatrix(1, 1)
    P00.setOption(PETSc.Mat.Option.SPD, True)
    P11.setOption(PETSc.Mat.Option.SPD, True)

    u_h, p_h = problem.solve()
    assert problem.solver.getConvergedReason() > 0
    # Because left-hand side operator is only assembled during solve
    # we can only test the null space at this point.
    assert nsp.test(problem.A)

    # Compute norms of the solution vectors
    norm_u = la.norm(u_h.x)
    norm_p = la.norm(p_h.x)
    if MPI.COMM_WORLD.rank == 0:
        print(f"(A) Norm of velocity coefficient vector (high-level nested, iterative): {norm_u}")
        print(f"(A) Norm of pressure coefficient vector (high-level nested, iterative): {norm_p}")

    return norm_u, norm_p


# ### Low-level nested matrix solver
#
# We assemble the bilinear form into a nested matrix `A`, and call the
# `assemble()` method to communicate shared entries in parallel. Rows
# and columns in `A` that correspond to degrees-of-freedom with
# Dirichlet boundary conditions will be zeroed by the assembler, and a
# value of 1 will be set on the diagonal for these rows.


def nested_iterative_solver_low_level():
    """Solve Stokes problem using nest matrices and an iterative solver.

    Used low-level DOLFINx routines.
    """
    # Assemble nested matrix operators
    A = assemble_matrix(a, bcs=bcs, kind="nest")
    A.assemble()

    # Create a nested matrix P to use as the preconditioner. The
    # top-left block of P is shared with the top-left block of A. The
    # bottom-right diagonal entry is assembled from the form a_p11:
    P11 = assemble_matrix(a_p11, [])
    P = PETSc.Mat().createNest([[A.getNestSubMatrix(0, 0), None], [None, P11]])
    P.assemble()

    A00 = A.getNestSubMatrix(0, 0)
    A00.setOption(PETSc.Mat.Option.SPD, True)

    P00, P11 = P.getNestSubMatrix(0, 0), P.getNestSubMatrix(1, 1)
    P00.setOption(PETSc.Mat.Option.SPD, True)
    P11.setOption(PETSc.Mat.Option.SPD, True)

    # Assemble right-hand side vector
    b = assemble_vector(L, kind="nest")

    # Modify ('lift') the RHS for Dirichlet boundary conditions
    bcs1 = bcs_by_block(extract_function_spaces(a, 1), bcs)
    apply_lifting(b, a, bcs=bcs1)

    # Sum contributions for vector entries that are shared across
    # parallel processes
    for b_sub in b.getNestSubVecs():
        b_sub.ghostUpdate(addv=PETSc.InsertMode.ADD, mode=PETSc.ScatterMode.REVERSE)

    # Set Dirichlet boundary condition values in the RHS vector
    bcs0 = bcs_by_block(extract_function_spaces(L), bcs)
    set_bc(b, bcs0)

    # Set the nullspace for pressure (since pressure is determined only
    # up to a constant)
    null_vec = create_vector(extract_function_spaces(L), "nest")
    null_vecs = null_vec.getNestSubVecs()
    null_vecs[0].set(0.0), null_vecs[1].set(1.0)
    null_vec.normalize()
    nsp = PETSc.NullSpace().create(vectors=[null_vec])
    assert nsp.test(A)
    A.setNullSpace(nsp)

    # Create a MINRES Krylov solver and a block-diagonal preconditioner
    # using PETSc's additive fieldsplit preconditioner
    ksp = PETSc.KSP().create(msh.comm)
    ksp.setOperators(A, P)
    ksp.setType("minres")
    ksp.setTolerances(rtol=1e-9)
    ksp.getPC().setType("fieldsplit")
    ksp.getPC().setFieldSplitType(PETSc.PC.CompositeType.ADDITIVE)

    # Define the matrix blocks in the preconditioner with the velocity
    # and pressure matrix index sets
    nested_IS = P.getNestISs()
    ksp.getPC().setFieldSplitIS(("u", nested_IS[0][0]), ("p", nested_IS[0][1]))

    # Set the preconditioners for each block. For the top-left
    # Laplace-type operator we use algebraic multigrid. For the
    # lower-right block we use a Jacobi preconditioner. By default, GAMG
    # will infer the correct near-nullspace from the matrix block size.
    ksp_u, ksp_p = ksp.getPC().getFieldSplitSubKSP()
    ksp_u.setType("preonly")
    ksp_u.getPC().setType("gamg")
    ksp_p.setType("preonly")
    ksp_p.getPC().setType("jacobi")

    # Create finite element {py:class}`Function <dolfinx.fem.Function>`s
    # for the velocity (on the space `V`) and for the pressure (on the
    # space `Q`). The vectors for `u` and `p` are combined to form a
    # nested vector and the system is solved.
    u, p = Function(V), Function(Q)
    x = PETSc.Vec().createNest([create_vector_wrap(u.x), create_vector_wrap(p.x)])
    ksp.solve(b, x)

    # Save solution to file in XDMF format for visualization, e.g. with
    # ParaView. Before writing to file, ghost values are updated using
    # `scatter_forward`.
    with XDMFFile(MPI.COMM_WORLD, "out_stokes/velocity.xdmf", "w") as ufile_xdmf:
        u.x.scatter_forward()
        P1 = element("Lagrange", msh.basix_cell(), 1, shape=(gdim,), dtype=default_real_type)
        u1 = Function(functionspace(msh, P1))
        u1.interpolate(u)
        ufile_xdmf.write_mesh(msh)
        ufile_xdmf.write_function(u1)

    with XDMFFile(MPI.COMM_WORLD, "out_stokes/pressure.xdmf", "w") as pfile_xdmf:
        p.x.scatter_forward()
        pfile_xdmf.write_mesh(msh)
        pfile_xdmf.write_function(p)

    # Compute norms of the solution vectors
    norm_u = la.norm(u.x)
    norm_p = la.norm(p.x)
    if MPI.COMM_WORLD.rank == 0:
        print(f"(A) Norm of velocity coefficient vector (low-level nested, iterative): {norm_u}")
        print(f"(A) Norm of pressure coefficient vector (low-level nested, iterative): {norm_p}")

    return norm_u, norm_p


# ### Monolithic block iterative solver
#
# We now solve the same Stokes problem, but using monolithic
# (non-nested) matrices. We first create a helper function for
# assembling the linear operators and the RHS vector.


def block_operators():
    """Block operators and block RHS vector for the Stokes problem."""
    # Assembler matrix operator, preconditioner and RHS vector into
    # single objects but preserving block structure
    A = assemble_matrix(a, bcs=bcs)
    A.assemble()
    P = assemble_matrix(a_p, bcs=bcs)
    P.assemble()

    b = assemble_vector(L, kind=PETSc.Vec.Type.MPI)
    bcs1 = bcs_by_block(extract_function_spaces(a, 1), bcs)
    apply_lifting(b, a, bcs=bcs1)
    b.ghostUpdate(addv=PETSc.InsertMode.ADD, mode=PETSc.ScatterMode.REVERSE)
    bcs0 = bcs_by_block(extract_function_spaces(L), bcs)
    set_bc(b, bcs0)

    # Set the nullspace for pressure (since pressure is determined only
    # up to a constant)
    null_vec = A.createVecLeft()
    offset = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    null_vec.array[offset:] = 1.0
    null_vec.normalize()
    nsp = PETSc.NullSpace().create(vectors=[null_vec])
    assert nsp.test(A)
    A.setNullSpace(nsp)

    return A, P, b


# The following function solves the Stokes problem using a
# block-diagonal preconditioner and monolithic PETSc matrices.


def block_iterative_solver():
    """Solve Stokes problem using blocked matrices and iterative solver."""
    # Assembler the operators and RHS vector
    A, P, b = block_operators()

    # Build PETSc index sets for each field (global dof indices for each
    # field)
    V_map = V.dofmap.index_map
    Q_map = Q.dofmap.index_map
    offset_u = V_map.local_range[0] * V.dofmap.index_map_bs + Q_map.local_range[0]
    offset_p = offset_u + V_map.size_local * V.dofmap.index_map_bs
    is_u = PETSc.IS().createStride(
        V_map.size_local * V.dofmap.index_map_bs, offset_u, 1, comm=msh.comm
    )
    is_p = PETSc.IS().createStride(Q_map.size_local, offset_p, 1, comm=msh.comm)

    # Create a MINRES Krylov solver and a block-diagonal preconditioner
    # using PETSc's additive fieldsplit preconditioner
    ksp = PETSc.KSP().create(msh.comm)
    ksp.setOperators(A, P)
    ksp.setTolerances(rtol=1e-9)
    ksp.setType("minres")
    ksp.getPC().setType("fieldsplit")
    ksp.getPC().setFieldSplitType(PETSc.PC.CompositeType.ADDITIVE)
    ksp.getPC().setFieldSplitIS(("u", is_u), ("p", is_p))

    # Configure velocity and pressure sub-solvers
    ksp_u, ksp_p = ksp.getPC().getFieldSplitSubKSP()
    ksp_u.setType("preonly")
    ksp_u.getPC().setType("gamg")
    ksp_p.setType("preonly")
    ksp_p.getPC().setType("jacobi")

    # The matrix A combined the vector velocity and scalar pressure
    # parts, hence has a block size of 1. Unlike the MatNest case, GAMG
    # cannot infer the correct near-nullspace from the matrix block
    # size. Therefore, we set block size on the top-left block of the
    # preconditioner so that GAMG can infer the appropriate near
    # nullspace.
    ksp.getPC().setUp()
    Pu, _ = ksp_u.getPC().getOperators()
    Pu.setBlockSize(msh.topology.dim)

    # Create a block vector (x) to store the full solution and solve
    x = A.createVecRight()
    ksp.solve(b, x)

    # Create Functions to split u and p
    u, p = Function(V), Function(Q)
    offset = V_map.size_local * V.dofmap.index_map_bs
    u.x.array[:offset] = x.array_r[:offset]
    p.x.array[: (len(x.array_r) - offset)] = x.array_r[offset:]

    # Compute the $L^2$ norms of the solution vectors
    norm_u, norm_p = la.norm(u.x), la.norm(p.x)
    if MPI.COMM_WORLD.rank == 0:
        print(f"(B) Norm of velocity coefficient vector (blocked, iterative): {norm_u}")
        print(f"(B) Norm of pressure coefficient vector (blocked, iterative): {norm_p}")

    return norm_u, norm_p


# ### Monolithic block direct solver
#
# We now solve the same Stokes problem again, but using monolithic
# (non-nested) matrices and a direct (LU) solver.


def block_direct_solver():
    """Solve Stokes problem using blocked matrices and a direct solver."""
    # Assembler the block operator and RHS vector
    A, _, b = block_operators()

    # Create a solver
    ksp = PETSc.KSP().create(msh.comm)
    ksp.setOperators(A)
    ksp.setType("preonly")

    # Set the solver type to MUMPS (LU solver) and configure MUMPS to
    # handle pressure nullspace
    pc = ksp.getPC()
    pc.setType("lu")
    use_superlu = PETSc.IntType == np.int64
    if PETSc.Sys().hasExternalPackage("mumps") and not use_superlu:
        pc.setFactorSolverType("mumps")
        pc.setFactorSetUpSolverType()
        pc.getFactorMatrix().setMumpsIcntl(icntl=24, ival=1)
        pc.getFactorMatrix().setMumpsIcntl(icntl=25, ival=0)
    else:
        pc.setFactorSolverType("superlu_dist")

    # Create a block vector (x) to store the full solution, and solve
    x = A.createVecLeft()
    ksp.solve(b, x)

    # Create Functions and scatter x solution
    u, p = Function(V), Function(Q)
    offset = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    u.x.array[:offset] = x.array_r[:offset]
    p.x.array[: (len(x.array_r) - offset)] = x.array_r[offset:]

    # Compute the $L^2$ norms of the u and p vectors
    norm_u, norm_p = la.norm(u.x), la.norm(p.x)
    if MPI.COMM_WORLD.rank == 0:
        print(f"(C) Norm of velocity coefficient vector (blocked, direct): {norm_u}")
        print(f"(C) Norm of pressure coefficient vector (blocked, direct): {norm_p}")

    return norm_u, norm_p


# ### Non-blocked direct solver
#
# We now solve the Stokes problem, but using monolithic matrix with the
# velocity and pressure degrees of freedom interleaved, i.e. without any
# u/p block structure in the assembled matrix. A direct (LU) solver is
# used.


def mixed_direct():
    """Solve Stokes problem using mixed matrix and a direct solver."""
    # Create the Taylot-Hood function space
    TH = mixed_element([P2, P1])
    W = functionspace(msh, TH)

    # No slip boundary condition
    W0 = W.sub(0)
    Q, _ = W0.collapse()
    noslip = Function(Q)
    facets = locate_entities_boundary(msh, 1, noslip_boundary)
    dofs = locate_dofs_topological((W0, Q), 1, facets)
    bc0 = dirichletbc(noslip, dofs, W0)

    # Driving velocity condition u = (1, 0) on top boundary (y = 1)
    lid_velocity = Function(Q)
    lid_velocity.interpolate(lid_velocity_expression)
    facets = locate_entities_boundary(msh, 1, lid)
    dofs = locate_dofs_topological((W0, Q), 1, facets)
    bc1 = dirichletbc(lid_velocity, dofs, W0)

    # Collect Dirichlet boundary conditions
    bcs = [bc0, bc1]

    # Define variational problem
    (u, p) = ufl.TrialFunctions(W)
    (v, q) = ufl.TestFunctions(W)
    f = Function(Q)
    a = form(
        (ufl.inner(ufl.grad(u), ufl.grad(v)) + ufl.inner(p, ufl.div(v)) + ufl.inner(ufl.div(u), q))
        * ufl.dx
    )
    L = form(ufl.inner(f, v) * ufl.dx)

    # Assemble LHS matrix and RHS vector
    A = assemble_matrix(a, bcs=bcs)
    A.assemble()
    b = assemble_vector(L)

    bcs1 = bcs_by_block(extract_function_spaces([[a]], 1), bcs)
    apply_lifting(b, [a], bcs=bcs1)
    b.ghostUpdate(addv=PETSc.InsertMode.ADD, mode=PETSc.ScatterMode.REVERSE)

    # Set Dirichlet boundary condition values in the RHS
    for bc in bcs:
        bc.set(b.array_w)

    # Create and configure solver
    ksp = PETSc.KSP().create(msh.comm)
    ksp.setOperators(A)
    ksp.setType("preonly")

    # Configure MUMPS to handle pressure nullspace
    pc = ksp.getPC()
    pc.setType("lu")
    use_superlu = PETSc.IntType == np.int64
    if PETSc.Sys().hasExternalPackage("mumps") and not use_superlu:
        pc.setFactorSolverType("mumps")
        pc.setFactorSetUpSolverType()
        pc.getFactorMatrix().setMumpsIcntl(icntl=24, ival=1)
        pc.getFactorMatrix().setMumpsIcntl(icntl=25, ival=0)
    else:
        pc.setFactorSolverType("superlu_dist")

    # Compute the solution
    U = Function(W)
    try:
        ksp.solve(b, U.x.petsc_vec)
    except PETSc.Error as e:
        if e.ierr == 92:
            print("The required PETSc solver/preconditioner is not available. Exiting.")
            print(e)
            exit(0)
        else:
            raise e

    # Split the mixed solution and collapse
    u, p = U.sub(0).collapse(), U.sub(1).collapse()

    # Compute norms
    norm_u, norm_p = la.norm(u.x), la.norm(p.x)
    if MPI.COMM_WORLD.rank == 0:
        print(f"(D) Norm of velocity coefficient vector (monolithic, direct): {norm_u}")
        print(f"(D) Norm of pressure coefficient vector (monolithic, direct): {norm_p}")

    return norm_u, norm_u


# Solve using LinearProblem class

norm_u_0, norm_p_0 = nested_iterative_solver_high_level()

# Solve using PETSc MatNest

norm_u_1, norm_p_1 = nested_iterative_solver_low_level()
np.testing.assert_allclose(norm_u_1, norm_u_0, rtol=1e-4)
np.testing.assert_allclose(norm_p_1, norm_p_0, rtol=1e-4)

# Solve using PETSc block matrices and an iterative solver

norm_u_2, norm_p_2 = block_iterative_solver()
np.testing.assert_allclose(norm_u_2, norm_u_0, rtol=1e-4)
np.testing.assert_allclose(norm_p_2, norm_p_0, rtol=1e-4)

# Solve using PETSc block matrices and an LU solver

norm_u_3, norm_p_3 = block_direct_solver()
np.testing.assert_allclose(norm_u_3, norm_u_0, rtol=1e-4)
np.testing.assert_allclose(norm_p_3, norm_p_0, rtol=1e-4)

# Solve using a non-blocked matrix and an LU solver

norm_u_4, norm_p_4 = mixed_direct()
np.testing.assert_allclose(norm_u_4, norm_u_0, rtol=1e-4)



# Copyright (C) 2019-2024 Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Tests for custom Python assemblers."""

import math
import time

from mpi4py import MPI

try:
    import numba

    from ffcx.codegeneration.utils import get_void_pointer
except ImportError:
    pass

import numpy as np
import pytest

import dolfinx
import ufl
from dolfinx.fem import Function, form, functionspace
from dolfinx.mesh import create_unit_square

cffi = pytest.importorskip("cffi")
cffi_support = pytest.importorskip("numba.core.typing.cffi_utils")
numba = pytest.importorskip("numba")

ffi = cffi.FFI()


# See https://github.com/numba/numba/issues/4036 for why we need 'sink'
@numba.njit
def sink(*args):
    pass


@numba.njit(fastmath=True)
def area(x0, x1, x2) -> float:
    """Compute the area of a triangle embedded in 2D from the three vertices."""
    a = (x1[0] - x2[0]) ** 2 + (x1[1] - x2[1]) ** 2
    b = (x0[0] - x2[0]) ** 2 + (x0[1] - x2[1]) ** 2
    c = (x0[0] - x1[0]) ** 2 + (x0[1] - x1[1]) ** 2
    return math.sqrt(2 * (a * b + a * c + b * c) - (a**2 + b**2 + c**2)) / 4.0


@numba.njit(fastmath=True)
def assemble_vector(b, mesh, dofmap, num_cells):
    """Assemble simple linear form over a mesh into the array b."""
    v, x = mesh
    q0, q1 = 1 / 3.0, 1 / 3.0
    for cell in range(num_cells):
        # FIXME: This assumes a particular geometry dof layout
        A = area(x[v[cell, 0]], x[v[cell, 1]], x[v[cell, 2]])
        b[dofmap[cell, 0]] += A * (1.0 - q0 - q1)
        b[dofmap[cell, 1]] += A * q0
        b[dofmap[cell, 2]] += A * q1


@numba.njit(parallel=True, fastmath=True)
def assemble_vector_parallel(b, v, x, dofmap_t_data, dofmap_t_offsets, num_cells):
    """Assemble simple linear form over a mesh into the array b using a parallel loop."""
    q0 = 1 / 3.0
    q1 = 1 / 3.0
    b_unassembled = np.empty((num_cells, 3), dtype=b.dtype)
    for cell in numba.prange(num_cells):
        # FIXME: This assumes a particular geometry dof layout
        A = area(x[v[cell, 0]], x[v[cell, 1]], x[v[cell, 2]])
        b_unassembled[cell, 0] = A * (1.0 - q0 - q1)
        b_unassembled[cell, 1] = A * q0
        b_unassembled[cell, 2] = A * q1

    # Accumulate values in RHS
    _b_unassembled = b_unassembled.reshape(num_cells * 3)
    for index in numba.prange(dofmap_t_offsets.shape[0] - 1):
        for p in range(dofmap_t_offsets[index], dofmap_t_offsets[index + 1]):
            b[index] += _b_unassembled[dofmap_t_data[p]]


@numba.njit(fastmath=True)
def assemble_vector_ufc(b, kernel, mesh, dofmap, num_cells, dtype):
    """Assemble provided FFCx/UFC kernel over a mesh into the array b."""
    v, x = mesh
    entity_local_index = np.array([0], dtype=np.intc)
    perm = np.array([0], dtype=np.uint8)
    geometry = np.zeros((3, 3), dtype=x.dtype)
    coeffs = np.zeros(1, dtype=dtype)
    constants = np.zeros(1, dtype=dtype)
    custom_data = np.zeros(1, dtype=np.int64)
    custom_data_ptr = get_void_pointer(custom_data)

    b_local = np.zeros(3, dtype=dtype)
    for cell in range(num_cells):
        # FIXME: This assumes a particular geometry dof layout
        for j in range(3):
            geometry[j] = x[v[cell, j], :]
        b_local.fill(0.0)
        kernel(
            ffi.from_buffer(b_local),
            ffi.from_buffer(coeffs),
            ffi.from_buffer(constants),
            ffi.from_buffer(geometry),
            ffi.from_buffer(entity_local_index),
            ffi.from_buffer(perm),
            custom_data_ptr,
        )
        for j in range(3):
            b[dofmap[cell, j]] += b_local[j]


@pytest.mark.parametrize(
    "dtype",
    [
        np.float32,
        np.float64,
        pytest.param(
            np.complex64,
            marks=[
                pytest.mark.xfail_win32_complex,
                pytest.mark.skipif(
                    cffi.__version_info__ > (1, 16, 99) and cffi.__version_info__ <= (2, 0, 0),
                    reason="bug in cffi 1.17.0/1 and 2.0.0 for complex",
                ),
            ],
        ),
        pytest.param(
            np.complex128,
            marks=[
                pytest.mark.xfail_win32_complex,
                pytest.mark.skipif(
                    cffi.__version_info__ > (1, 16, 99) and cffi.__version_info__ <= (2, 0, 0),
                    reason="bug in cffi 1.17.0/1 and 2.0.0 for complex",
                ),
            ],
        ),
    ],
)
def test_custom_mesh_loop_rank1(dtype):
    mesh = create_unit_square(MPI.COMM_WORLD, 64, 64, dtype=dtype(0).real.dtype)
    V = functionspace(mesh, ("Lagrange", 1))

    # Unpack mesh and dofmap data
    num_owned_cells = mesh.topology.index_map(mesh.topology.dim).size_local
    x_dofs = mesh.geometry.dofmaps[0]
    x = mesh.geometry.x
    dofmap = V.dofmap.list

    # Assemble with pure Numba function (two passes, first will include
    # JIT overhead)
    b0 = Function(V, dtype=dtype)
    for i in range(2):
        b = b0.x.array
        b[:] = 0.0
        start = time.time()
        assemble_vector(b, (x_dofs, x), dofmap, num_owned_cells)
        end = time.time()
        print(f"Time (numba, pass {i}): {end - start}")
    b0.x.scatter_reverse(dolfinx.la.InsertMode.add)
    b0sum = np.sum(b0.x.array[: b0.x.index_map.size_local * b0.x.block_size])
    assert mesh.comm.allreduce(b0sum, op=MPI.SUM) == pytest.approx(1.0)

    # NOTE: Parallel (threaded) Numba can cause problems with MPI
    # Assemble with pure Numba function using parallel loop (two passes,
    # first will include JIT overhead)
    # from dolfinx.fem import transpose_dofmap
    # dofmap_t = transpose_dofmap(V.dofmap.list, num_owned_cells)
    # btmp = Function(V)
    # for i in range(2):
    #     b = btmp.x.array
    #     b[:] = 0.0
    #     start = time.time()
    #     assemble_vector_parallel(b, x_dofs, x, dofmap_t.array, dofmap_t.offsets, num_owned_cells)
    #     end = time.time()
    #     print("Time (numba parallel, pass {}): {}".format(i, end - start))
    # btmp.x.scatter_reverse(dolfinx.la.InsertMode.add)
    # assert (btmp.x.array - b0.x.array).norm() == pytest.approx(0.0)

    # Test against generated code and general assembler
    v = ufl.TestFunction(V)
    L = ufl.inner(1.0, v) * ufl.dx
    Lf = form(L, dtype=dtype)
    start = time.time()
    b1 = dolfinx.fem.assemble_vector(Lf)
    end = time.time()
    print("Time (C++, pass 0):", end - start)

    b1.array[:] = 0
    start = time.time()
    dolfinx.fem.assemble_vector(b1.array, Lf)
    end = time.time()
    print("Time (C++, pass 1):", end - start)
    b1.scatter_reverse(dolfinx.la.InsertMode.add)
    assert np.linalg.norm(b1.array - b0.x.array) == pytest.approx(0.0, abs=1.0e-8)

    # Assemble using generated tabulate_tensor kernel and Numba
    # assembler
    b3 = Function(V, dtype=dtype)
    ufcx_form, _module, _code = dolfinx.jit.ffcx_jit(
        mesh.comm, L, form_compiler_options={"scalar_type": dtype}
    )

    # Get the one and only kernel
    kernel = getattr(ufcx_form.form_integrals[0], f"tabulate_tensor_{np.dtype(dtype).name}")

    for i in range(2):
        b = b3.x.array
        b[:] = 0.0
        start = time.time()
        assemble_vector_ufc(b, kernel, (x_dofs, x), dofmap, num_owned_cells, dtype)
        end = time.time()
        print(f"Time (numba/cffi, pass {i}): {end - start}")
    b3.x.scatter_reverse(dolfinx.la.InsertMode.add)
    assert np.linalg.norm(b3.x.array - b0.x.array) == pytest.approx(0.0, abs=1e-8)



from mpi4py import MPI

import numpy as np
import pytest

import basix.ufl
import dolfinx
import ufl


@pytest.mark.parametrize("L", [0.1, 0.2, 0.3])
@pytest.mark.parametrize("H", [1.3, 0.8, 0.2])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.triangle, dolfinx.mesh.CellType.quadrilateral]
)
@pytest.mark.parametrize("dtype", [np.float64, np.float32])
def test_real_function_space_mass(L, H, cell_type, dtype):
    """Test that real space mass matrix is the same as assembling the volume."""
    mesh = dolfinx.mesh.create_rectangle(
        MPI.COMM_WORLD, [[0.0, 0.0], [L, H]], [7, 9], cell_type, dtype=dtype
    )

    el = basix.ufl.real_element(mesh.basix_cell(), dtype=dtype)
    V = dolfinx.fem.functionspace(mesh, el)
    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    a = ufl.inner(u, v) * ufl.dx

    A = dolfinx.fem.assemble_matrix(dolfinx.fem.form(a, dtype=dtype), bcs=[])
    A.scatter_reverse()
    tol = 100 * np.finfo(dtype).eps
    cell_map = mesh.topology.index_map(mesh.topology.dim)
    if cell_map.size_local + cell_map.num_ghosts > 0:
        assert len(A.data) == 1
        if cell_map.local_range[0] == 0:
            assert np.isclose(A.data[0], L * H, atol=tol)
    else:
        assert len(A.data) == 0
        assert len(V.dofmap.list.flatten()) == 0
        assert V.dofmap.index_map.size_local == 0
        assert V.dofmap.index_map.num_ghosts == 1


@pytest.mark.parametrize("dtype", [np.float64, np.float32])
@pytest.mark.parametrize(
    "cell_type", [dolfinx.mesh.CellType.tetrahedron, dolfinx.mesh.CellType.hexahedron]
)
def test_real_function_space_vector(cell_type, dtype):
    """Test assembling with real space test function is equal to assembling with a constant."""
    mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 2, 3, 5, cell_type, dtype=dtype)

    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 3))
    v = ufl.TrialFunction(V)

    el = basix.ufl.real_element(mesh.basix_cell(), dtype=dtype)
    R = dolfinx.fem.functionspace(mesh, el)
    u = ufl.TestFunction(R)
    a_R = ufl.inner(u, v) * ufl.dx
    form_rhs = dolfinx.fem.form(a_R, dtype=dtype)

    A_R = dolfinx.fem.assemble_matrix(form_rhs, bcs=[])
    A_R.scatter_reverse()

    L = ufl.inner(ufl.constantvalue.IntValue(1), v) * ufl.dx
    form_lhs = dolfinx.fem.form(L, dtype=dtype)
    b = dolfinx.fem.assemble_vector(form_lhs)
    b.scatter_reverse(dolfinx.la.InsertMode.add)
    b.scatter_forward()

    row_map = A_R.index_map(0)
    num_local_rows = row_map.size_local
    num_dofs = V.dofmap.index_map.size_local * V.dofmap.index_map_bs
    tol = 100 * np.finfo(dtype).eps
    if MPI.COMM_WORLD.rank == 0:
        assert num_local_rows == 1
        num_dofs_global = V.dofmap.index_map.size_global * V.dofmap.index_map_bs
        assert A_R.indptr[1] - A_R.indptr[0] == num_dofs_global
        np.testing.assert_allclose(A_R.indices, np.arange(num_dofs_global))
        np.testing.assert_allclose(b.array[:num_dofs], A_R.data[:num_dofs], atol=tol)
    else:
        assert num_local_rows == 0


@pytest.mark.parametrize(
    "ftype, stype",
    [
        pytest.param(np.float32, np.complex64, marks=pytest.mark.xfail_win32_complex),
        pytest.param(np.float64, np.complex128, marks=pytest.mark.xfail_win32_complex),
    ],
)
def test_complex_real_space(ftype, stype):
    mesh = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 13, dtype=ftype)

    val = (2 + 3j, -4 + 5j)
    value_shape = (2,)
    el = basix.ufl.real_element(mesh.basix_cell(), value_shape=value_shape, dtype=ftype)
    R = dolfinx.fem.functionspace(mesh, el)
    u = dolfinx.fem.Function(R, dtype=stype)
    u.x.array[0] = val[0]
    u.x.array[1] = val[1]
    u.x.scatter_forward()

    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 1, value_shape))
    v = ufl.TestFunction(V)
    L = ufl.inner(u, v) * ufl.dx

    b = dolfinx.fem.assemble_vector(dolfinx.fem.form(L, dtype=stype))
    b.scatter_reverse(dolfinx.la.InsertMode.add)
    b.scatter_forward()
    const = dolfinx.fem.Constant(mesh, stype(val))
    L_const = ufl.inner(const, v) * ufl.dx
    b_const = dolfinx.fem.assemble_vector(dolfinx.fem.form(L_const, dtype=stype))
    b_const.scatter_reverse(dolfinx.la.InsertMode.add)
    b_const.scatter_forward()

    tol = 100 * np.finfo(stype).eps
    np.testing.assert_allclose(b.array, b_const.array, atol=tol)



# Copyright (C) 2025 Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from mpi4py import MPI

import numpy as np
import pytest


@pytest.mark.parametrize(
    "marker_mode",
    [
        pytest.param(0, marks=pytest.mark.xfail(raises=RuntimeError)),
        pytest.param(1, marks=pytest.mark.xfail(raises=RuntimeError)),
        2,
        pytest.param(3, marks=pytest.mark.xfail(raises=RuntimeError)),
    ],
)
def test_physical_tags(marker_mode):
    """Test that we catch partially tagged meshes and not tagged
    meshes as errors.
    """
    gmsh = pytest.importorskip("gmsh")

    from dolfinx.io import gmsh as gmshio

    gmsh.initialize()

    def gmsh_tet_model(order):
        gmsh.option.setNumber("General.Terminal", 0)
        model = gmsh.model()
        comm = MPI.COMM_WORLD
        if comm.rank == 0:
            model.add("Sphere minus box")
            model.setCurrent("Sphere minus box")
            model.occ.addSphere(0, 0, 0, 1)
            model.occ.addSphere(2, 2, 2, 0.3)
            model.occ.synchronize()
            volume_entities = [model[1] for model in model.getEntities(3)]
            volume_entities = volume_entities[:marker_mode]
            for i, entity in enumerate(volume_entities):
                model.addPhysicalGroup(3, [entity], tag=i)
            if marker_mode == 3:  # Check duplicate marker error
                model.addPhysicalGroup(3, [entity], tag=10)
            model.mesh.generate(3)
            gmsh.option.setNumber("General.Terminal", 1)
            model.mesh.setOrder(order)
            gmsh.option.setNumber("General.Terminal", 0)

        mesh_data = gmshio.model_to_mesh(model, comm, 0)
        return mesh_data.mesh, mesh_data.cell_tags

    msh, cell_tags = gmsh_tet_model(1)
    gdim = msh.geometry.dim
    assert msh.geometry.cmaps[0].degree == 1
    assert msh.geometry.dim == gdim
    local_values = np.unique(cell_tags.values)
    all_values = np.unique(np.hstack(msh.comm.allgather(local_values)))
    assert len(all_values) == 2

    gmsh.finalize()



# Copyright (C) 2021-2024 Garth N. Wells and Paul T. Kühner
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Graph representations and operations on graphs."""

from typing import Generic

import numpy as np
import numpy.typing as npt

from dolfinx import cpp as _cpp
from dolfinx.cpp.graph import partitioner
from dolfinx.typing import Index

# Import graph partitioners, which may or may not be available
# (dependent on build configuration)
try:
    from dolfinx.cpp.graph import partitioner_scotch  # noqa
except ImportError:
    pass
try:
    from dolfinx.cpp.graph import partitioner_parmetis  # noqa
except ImportError:
    pass
try:
    from dolfinx.cpp.graph import partitioner_kahip  # noqa
except ImportError:
    pass


__all__ = [
    "AdjacencyList",
    "adjacencylist",
    "comm_graph",
    "comm_graph_data",
    "comm_to_json",
    "partitioner",
]


class AdjacencyList(Generic[Index]):
    """Adjacency list representation of a graph."""

    _cpp_object: (
        _cpp.graph.AdjacencyList_int32
        | _cpp.graph.AdjacencyList_int64
        | _cpp.graph.AdjacencyList_int_sizet_int8__int32_int32
    )

    def __init__(
        self,
        g: (
            _cpp.graph.AdjacencyList_int32
            | _cpp.graph.AdjacencyList_int64
            | _cpp.graph.AdjacencyList_int_sizet_int8__int32_int32
        ),
    ):
        """Creates a Python wrapper for the exported adjacency list class.

        Note:
            Do not use this constructor directly. Instead use
            :func:`adjacencylist`.

        Args:
            g: The underlying cpp instance that this object will wrap.
        """
        self._cpp_object = g

    def __repr__(self):
        """String representation of the adjacency list."""
        return self._cpp_object.__repr__()

    def links(self, node: int) -> npt.NDArray[Index]:
        """Retrieve the links of a node.

        Note:
            This is available only for adjacency lists with no
            additional link (edge) data.

        Args:
            node: Node to retrieve the connectivity of.

        Returns:
            Neighbors of the node.
        """
        return self._cpp_object.links(node)

    @property
    def array(self) -> npt.NDArray[Index]:
        """Array representation of the adjacency list.

        Note:
            This is available only for adjacency lists with no
            additional link (edge) data.

        Returns:
            Flattened array representation of the adjacency list.
        """
        return self._cpp_object.array

    @property
    def offsets(self) -> npt.NDArray[np.int32]:
        """Offsets for each node in the :func:`array`.

        Returns:
            Array of indices with shape `(num_nodes+1)`.
        """
        return self._cpp_object.offsets

    @property
    def num_nodes(self) -> np.int32:
        """Number of nodes in the adjacency list.

        Returns:
            Number of nodes.
        """
        return self._cpp_object.num_nodes


def adjacencylist(
    data: npt.NDArray[Index], offsets: npt.NDArray[np.int32] | None = None
) -> AdjacencyList[Index]:
    """Create an :class:`AdjacencyList` for `int32` or `int64` datasets.

    Args:
        data: The adjacency array. If the array is one-dimensional,
            offsets should be supplied. If the array is two-dimensional
            the number of edges per node is the second dimension.
        offsets: The offsets array with the number of edges per node.

    Returns:
        An adjacency list.
    """
    # TODO: Switch to np.isdtype(data.dtype, np.int32) once numpy >= 2.0 is
    # enforced
    if data.dtype == np.int32:
        cpp_t = _cpp.graph.AdjacencyList_int32
    elif data.dtype == np.int64:
        cpp_t = _cpp.graph.AdjacencyList_int64
    else:
        raise TypeError("Data type for adjacency list not supported.")

    cpp_object = cpp_t(data, offsets) if offsets is not None else cpp_t(data)
    return AdjacencyList(cpp_object)


def comm_graph(map: _cpp.common.IndexMap, root: int = 0) -> AdjacencyList:
    """Build a parallel communication graph from an index map.

    The communication graph is a directed graph that represents the
    communication pattern for a distributed array, and specifically the
    forward scatter operation where the values for owned indices are
    sent to ghosting ranks. The graph is built from an index map, which
    describes the local and ghosted indices of the array.

    Edges in the graph represent communication from the owning rank to
    ranks that ghost the data. The edge data holds the (0) target node,
    (1) edge weight, and (2) an indicator for whether the sending and
    receiving ranks share memory (``local==1``) or if the ranks do not
    share memory (``local==0``). The node data holds the local size
    (number of owned indices) and the number of ghost indices.

    The graph can be processed using :func:`comm_graph` to build data
    structures that can be used to build a `NetworkX
    <https://networkx.org/>`_ directed graph.

    Note:
        This function is collective across all MPI ranks. The
        communication graph is returned on the `root` rank. All other
        ranks return an empty graph

    Args:
        map: Index map to build the communication graph from.
        root: Rank that will return the communication graph. Other ranks
            return an empty graph.

    Returns:
        An adjacency list representing the communication graph.
    """
    return AdjacencyList(_cpp.graph.comm_graph(map))


def comm_graph_data(
    graph: AdjacencyList,
) -> tuple[list[tuple[int, int, dict[str, int]]], list[tuple[int, dict[str, int]]]]:
    """Build communication graph data for use with `NetworkX <https://networkx.org/>`_.

    Args:
        graph: Communication graph to build data from. Normally created
            by :func:`comm_graph`.

    Returns:
        A tuple of two lists. The first list contains the edge data,
        where an edge is a `(nodeID_0, nodeID_1, dict)` tuple, where
        `dict` holds edge data. The second list hold node data, where a
        node is a `(nodeID, dict)` tuple, where `dict` holds node data.
    """
    return _cpp.graph.comm_graph_data(graph._cpp_object)


def comm_to_json(graph: AdjacencyList) -> str:
    """Build and JSON string from a communication graph.

    The JSON string can be used to construct a `NetworkX
    <https://networkx.org/>`_ graph. This is helpful for cases where a
    simulation is executed and the graph data is written to file as a
    JSON string for later analysis.

    Args:
        graph: The communication graph to convert. Normally created by
            calling :meth:`comm_graph`.

    Returns:
        A JSON string representing the communication graph.
    """
    return _cpp.graph.comm_to_json(graph._cpp_object)



# Copyright (C) 2025 Joseph P. Dean and Chris Richardson
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

# TODO Clean up these tests

from mpi4py import MPI

import numpy as np
import pytest

import dolfinx
from dolfinx.cpp.log import set_thread_name
from dolfinx.cpp.mesh import (
    Mesh_float32,
    Mesh_float64,
    compute_mixed_cell_pairs,
    create_cell_partitioner,
    create_geometry,
    create_mesh,
    create_topology,
    locate_entities,
)
from dolfinx.fem import coordinate_element
from dolfinx.log import LogLevel, set_log_level
from dolfinx.mesh import CellType, GhostMode, Mesh, Topology, create_unit_cube


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_mixed_topology_mesh(dtype):
    set_log_level(LogLevel.INFO)

    cells = [[0, 1, 2, 1, 2, 3], [2, 3, 4, 5]]
    orig_index = [[0, 1], [2]]
    ghost_owners = [[], []]
    boundary_vertices = []

    topology = Topology(
        create_topology(
            MPI.COMM_SELF,
            [CellType.triangle, CellType.quadrilateral],
            cells,
            orig_index,
            ghost_owners,
            boundary_vertices,
            0,
        )
    )

    maps = topology.index_maps(topology.dim)
    assert len(maps) == 2

    # Two triangles and one quad
    assert maps[0].size_local == 2
    assert maps[1].size_local == 1

    # Six vertices in map
    map0 = topology.index_maps(0)
    assert len(map0) == 1
    assert map0[0].size_local == 6

    entity_types = topology.entity_types
    assert len(entity_types[0]) == 1

    topology.create_entities(1, dolfinx.hardware_concurrency())
    entity_types = topology.entity_types
    assert len(entity_types[1]) == 1
    assert CellType.interval in entity_types[1]

    entity_types = topology.entity_types
    assert len(entity_types[2]) == 2

    # Two triangle cells
    assert entity_types[2][0] == CellType.triangle
    assert topology.connectivity((2, 0), (0, 0)).num_nodes == 2

    # One quadrlilateral cell
    assert entity_types[2][1] == CellType.quadrilateral
    assert topology.connectivity((2, 1), (0, 0)).num_nodes == 1

    # Create dofmaps for Geometry
    tri = coordinate_element(CellType.triangle, 1, dtype=dtype)
    quad = coordinate_element(CellType.quadrilateral, 1, dtype=dtype)
    nodes = np.array([0, 1, 2, 3, 4, 5], dtype=np.int64)
    xdofs = np.array([0, 1, 2, 1, 2, 3, 2, 3, 4, 5], dtype=np.int64)
    x = np.array([0.0, 0.0, 1.0, 0.0, 0.0, 1.0, 1.0, 1.0, 2.0, 1.0, 2.0, 0.0], dtype=dtype)
    geom = create_geometry(
        topology._cpp_object, [tri._cpp_object, quad._cpp_object], nodes, xdofs, x, 2
    )
    print(geom.x)
    print(geom.index_map().size_local)
    print(geom.dofmaps[0])
    print(geom.dofmaps[1])
    set_log_level(LogLevel.WARNING)


def test_mixed_topology_mesh_3d():
    # Mesh = 2 tets, 1 prism, 1 hex, joined.
    cells = [[0, 1, 2, 3, 1, 2, 3, 4], [2, 3, 4, 5, 6, 7], [3, 4, 6, 7, 8, 9, 10, 11]]
    orig_index = [[0, 1], [2], [3]]
    ghost_owners = [[], [], []]
    boundary_vertices = []

    topology = Topology(
        create_topology(
            MPI.COMM_SELF,
            [CellType.tetrahedron, CellType.prism, CellType.hexahedron],
            cells,
            orig_index,
            ghost_owners,
            boundary_vertices,
            0,
        )
    )

    entity_types = topology.entity_types
    assert len(entity_types[0]) == 1

    topology.create_entities(1, dolfinx.hardware_concurrency())
    entity_types = topology.entity_types
    assert len(entity_types[1]) == 1

    topology.create_entities(2, dolfinx.hardware_concurrency())
    entity_types = topology.entity_types
    assert len(entity_types[2]) == 2

    assert len(entity_types[3]) == 3

    # Create triangle and quadrilateral facets
    topology.create_entities(2, dolfinx.hardware_concurrency())

    qi = topology.entity_types[2].index(CellType.quadrilateral)
    ti = topology.entity_types[2].index(CellType.triangle)

    # Tet -> quad
    with pytest.raises(RuntimeError):
        t = topology.connectivity((3, 0), (2, qi))
    # Tet -> triangle
    t = topology.connectivity((3, 0), (2, ti))
    assert t.num_nodes == 2
    assert len(t.links(0)) == 4

    # Prism -> quad
    t = topology.connectivity((3, 1), (2, qi))
    assert t.num_nodes == 1
    assert len(t.links(0)) == 3
    # Prism -> triangle
    t = topology.connectivity((3, 1), (2, ti))
    assert t.num_nodes == 1
    assert len(t.links(0)) == 2

    # Hex -> quad
    t = topology.connectivity((3, 2), (2, qi))
    assert t.num_nodes == 1
    assert len(t.links(0)) == 6
    # Hex -> triangle
    with pytest.raises(RuntimeError):
        t = topology.connectivity((3, 2), (2, ti))

    # Quad -> vertex
    t = topology.connectivity((2, qi), (0, 0))
    assert t.num_nodes == 8
    assert len(t.links(0)) == 4

    # Triangle -> vertex
    t = topology.connectivity((2, ti), (0, 0))
    assert t.num_nodes == 8
    assert len(t.links(0)) == 3

    topology.create_connectivity(2, 1)

    # Quad -> edge
    t = topology.connectivity((2, qi), (1, 0))
    assert t.num_nodes == 8
    assert len(t.links(0)) == 4

    # Tri -> edge
    t = topology.connectivity((2, ti), (1, 0))
    assert t.num_nodes == 8
    assert len(t.links(0)) == 3

    topology.create_connectivity(2, 3)
    # Quad -> prism
    t = topology.connectivity((2, qi), (3, 1))
    assert t.num_nodes == 8
    assert t.array.size == 3

    # Quad -> hex
    t = topology.connectivity((2, qi), (3, 2))
    assert t.num_nodes == 8
    assert t.array.size == 6

    # Tri -> tet
    t = topology.connectivity((2, ti), (3, 0))
    assert t.num_nodes == 8
    assert t.array.size == 8

    # Tri -> prism
    t = topology.connectivity((2, ti), (3, 1))
    assert t.num_nodes == 8
    assert t.array.size == 2


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_parallel_mixed_mesh(dtype):
    rank = MPI.COMM_WORLD.Get_rank()

    # Two triangles and one quadrilateral
    tri = np.array([0, 1, 4, 0, 3, 4], dtype=np.int64)
    quad = np.array([1, 4, 2, 5], dtype=np.int64)
    # cells with global indexing
    cells = [[t + 3 * rank for t in tri], [q + 3 * rank for q in quad]]
    orig_index = [[3 * rank, 1 + 3 * rank], [2 + 3 * rank]]
    # No ghosting
    ghost_owners = [[], []]
    # All vertices are on boundary
    boundary_vertices = [3 * rank + i for i in range(6)]

    topology = create_topology(
        MPI.COMM_WORLD,
        [CellType.triangle, CellType.quadrilateral],
        cells,
        orig_index,
        ghost_owners,
        boundary_vertices,
        0,
    )

    # Cell types appear in order as in create_topology
    assert topology.entity_types[2][0] == CellType.triangle
    assert topology.entity_types[2][1] == CellType.quadrilateral

    size = MPI.COMM_WORLD.Get_size()
    assert topology.index_maps(2)[0].size_global == size * 2
    assert topology.index_maps(2)[1].size_global == size

    # Create dofmaps for Geometry
    tri = coordinate_element(CellType.triangle, 1, dtype=dtype)
    quad = coordinate_element(CellType.quadrilateral, 1, dtype=dtype)
    nodes = np.arange(6, dtype=np.int64) + 3 * rank
    xdofs = np.array([0, 1, 4, 0, 3, 4, 1, 4, 2, 5], dtype=np.int64) + 3 * rank
    x = np.array(
        [[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [0.0, 1.0], [1.0, 1.0], [2.0, 1.0]], dtype=dtype
    )
    x[:, 1] += 1.0 * rank

    set_log_level(LogLevel.INFO)
    set_thread_name(str(rank))
    geom = create_geometry(
        topology, [tri._cpp_object, quad._cpp_object], nodes, xdofs, x.flatten(), 2
    )

    assert len(geom.dofmaps[0]) == 2
    assert len(geom.dofmaps[1]) == 1

    mesh_t = Mesh_float64 if dtype == np.float64 else Mesh_float32
    mesh = mesh_t(MPI.COMM_WORLD, topology, geom)
    tri = mesh.topology.connectivity((2, 0), (0, 0))
    quad = mesh.topology.connectivity((2, 1), (0, 0))
    assert len(tri.array) == 6
    assert len(quad.array) == 4
    w = list(tri.array) + list(quad.array)
    assert max(w) == 5
    assert min(w) == 0
    print(tri.array, quad.array)

    set_log_level(LogLevel.WARNING)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_create_entities(dtype):
    mesh = create_unit_cube(
        MPI.COMM_WORLD, 2, 2, 2, CellType.prism, ghost_mode=GhostMode.none, dtype=dtype
    )

    # Make triangle and quadrilateral facets
    mesh.topology.create_entities(2, dolfinx.hardware_concurrency())

    assert len(mesh.topology.entity_types[2]) == 2
    qi = mesh.topology.entity_types[2].index(CellType.quadrilateral)
    ti = mesh.topology.entity_types[2].index(CellType.triangle)
    assert qi != ti

    cell_quad = mesh.topology.connectivity((3, 0), (2, qi))
    cell_tri = mesh.topology.connectivity((3, 0), (2, ti))

    assert MPI.COMM_WORLD.allreduce(cell_quad.num_nodes) == 16
    assert len(cell_quad.links(0)) == 3
    assert MPI.COMM_WORLD.allreduce(cell_tri.num_nodes) == 16
    assert len(cell_tri.links(0)) == 2

    quad_v = mesh.topology.connectivity((2, qi), (0, 0))
    tri_v = mesh.topology.connectivity((2, ti), (0, 0))
    ims = mesh.topology._cpp_object.index_maps(2)
    assert len(ims) == 2
    assert ims[qi].size_global == 32
    assert len(quad_v.links(0)) == 4
    assert ims[ti].size_global == 24
    assert len(tri_v.links(0)) == 3

    mesh.topology.create_entities(1, dolfinx.hardware_concurrency())
    # 9 edges on each prism
    cell_edge = mesh.topology.connectivity((3, 0), (1, 0))
    assert cell_edge.links(0).size == 9

    # Triangle and quad to prism (facet->cell)
    mesh.topology.create_connectivity(2, 3)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_locate_entities(dtype):
    # Create a unit cube mesh with one hex and two wedges
    if MPI.COMM_WORLD.rank == 0:
        hexes = np.array([0, 1, 3, 4, 6, 7, 9, 10], dtype=np.int64)
        wedges = np.array([1, 2, 4, 7, 8, 10, 2, 4, 5, 8, 10, 11], dtype=np.int64)
        cells = [hexes, wedges]
        geom = np.array(
            [
                [0.0, 0.0, 0.0],
                [0.5, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.5, 1.0, 0.0],
                [1.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.5, 0.0, 1.0],
                [1.0, 0.0, 1.0],
                [0.0, 1.0, 1.0],
                [0.5, 1.0, 1.0],
                [1.0, 1.0, 1.0],
            ],
            dtype=dtype,
        )
    else:
        cells = [np.array([], dtype=np.int64), np.array([], dtype=np.int64)]
        geom = np.array([], dtype=dtype)

    part = create_cell_partitioner(GhostMode.none, 2)
    hexahedron = coordinate_element(CellType.hexahedron, 1, dtype=dtype)
    prism = coordinate_element(CellType.prism, 1, dtype=dtype)
    comm = MPI.COMM_WORLD
    max_cells_per_facet = 2
    mesh = create_mesh(
        comm, cells, [hexahedron._cpp_object, prism._cpp_object], geom, part, max_cells_per_facet
    )

    fdim = mesh.topology.dim - 1

    def top(x):
        return np.isclose(x[2], 1.0)

    def front(x):
        return np.isclose(x[1], 0.0)

    facet_types = mesh.topology.entity_types[fdim]
    quad_idx = facet_types.index(CellType.quadrilateral)
    tri_idx = facet_types.index(CellType.triangle)

    # Should have one quadrilateral on top
    facets = locate_entities(mesh, fdim, top, quad_idx)
    assert MPI.Comm.allreduce(comm, len(facets), MPI.SUM) == 1

    # Should have two triangles on top
    facets = locate_entities(mesh, fdim, top, tri_idx)
    assert MPI.Comm.allreduce(comm, len(facets), MPI.SUM) == 2

    # Should have two quadrilaterals at the front
    facets = locate_entities(mesh, fdim, front, quad_idx)
    assert MPI.Comm.allreduce(comm, len(facets), MPI.SUM) == 2

    # Should have no triagles at the front
    facets = locate_entities(mesh, fdim, front, tri_idx)
    assert MPI.Comm.allreduce(comm, len(facets), MPI.SUM) == 0


def test_mixed_cell_pairs(mixed_topology_mesh):
    mesh = Mesh(mixed_topology_mesh, None)
    mesh.topology.create_entities(2, dolfinx.hardware_concurrency())
    mesh.topology.create_connectivity(2, 3)
    cell_types = mesh.topology.entity_types[3]
    facet_types = mesh.topology.entity_types[2]
    print(cell_types, facet_types)

    # For each facet type
    for f, ft in enumerate(facet_types):
        cell_pairs = compute_mixed_cell_pairs(mesh.topology._cpp_object, ft)
        for i, cti in enumerate(cell_types):
            for j, ctj in enumerate(cell_types):
                idx = i * len(cell_types) + j
                num_conns = len(cell_pairs[idx]) // 4
                print(f"Connectivity ({ft}) from {cti} to {ctj} : {num_conns}")
                if len(cell_pairs[idx]) > 0:
                    connection = np.array(cell_pairs[idx]).reshape((num_conns, -1))
                    f0 = mesh.topology.connectivity((3, i), (2, f))
                    f1 = mesh.topology.connectivity((3, j), (2, f))
                    for row in connection:
                        assert f0.links(row[0])[row[1]] == f1.links(row[2])[row[3]]



# Copyright (C) 2024-2025 Chris Richardson and Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from mpi4py import MPI

import numpy as np
import pytest

import dolfinx
import ufl
from dolfinx.io.vtkhdf import read_mesh, write_cell_data, write_mesh, write_point_data
from dolfinx.mesh import CellType, Mesh, create_unit_cube, create_unit_square


def test_read_write_vtkhdf_mesh2d():
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5, dtype=np.float32)
    write_mesh("example2d.vtkhdf", mesh)
    mesh2 = read_mesh(MPI.COMM_WORLD, "example2d.vtkhdf", np.float32)
    assert mesh2.geometry.x.dtype == np.float32
    mesh2 = read_mesh(MPI.COMM_WORLD, "example2d.vtkhdf", np.float64)
    assert mesh2.geometry.x.dtype == np.float64
    assert mesh.topology.index_map(2).size_global == mesh2.topology.index_map(2).size_global


def test_read_write_vtkhdf_mesh3d():
    mesh = create_unit_cube(MPI.COMM_WORLD, 5, 5, 5, cell_type=CellType.prism)
    write_mesh("example3d.vtkhdf", mesh)
    mesh2 = read_mesh(MPI.COMM_WORLD, "example3d.vtkhdf")

    assert mesh.topology.index_map(3).size_global == mesh2.topology.index_map(3).size_global


def test_read_write_mixed_topology(mixed_topology_mesh):
    mesh = Mesh(mixed_topology_mesh, None)
    write_mesh("mixed_mesh.vtkhdf", mesh)

    mesh2 = read_mesh(MPI.COMM_WORLD, "mixed_mesh.vtkhdf", np.float64)
    for t in mesh2.topology.entity_types[-1]:
        assert t in mesh.topology.entity_types[-1]


def test_read_write_higher_order():
    # Create a simple, 2 cell mesh consisting of a second order quadrilateral and
    # a second order triangle.
    geom = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
            [0.5, 0],
            [1, 0.5],
            [0.5, 1],
            [0, 0.5],
            [0.5, 0.5],
            [2.0, 0],
            [1.5, -0.2],
            [1.5, 0.6],
        ],
        dtype=np.float64,
    )
    # Nodes ordered as VTK
    if MPI.COMM_WORLD.rank == 0:
        topology_quad = np.array([[0, 1, 2, 3, 4, 5, 6, 7, 8]], dtype=np.int64)
        topology_tri = np.array([[1, 9, 2, 10, 11, 5]], dtype=np.int64)

    else:
        topology_quad = np.empty((0, 9), dtype=np.int64)
        topology_tri = np.empty((0, 6), dtype=np.int64)

    quad_perm = dolfinx.io.utils.cell_perm_vtk(dolfinx.mesh.CellType.quadrilateral, 9)
    tri_perm = dolfinx.io.utils.cell_perm_vtk(dolfinx.mesh.CellType.triangle, 6)
    topology_quad = topology_quad[:, quad_perm]
    topology_tri = topology_tri[:, tri_perm]

    cells_np = [topology_quad.flatten(), topology_tri.flatten()]
    coordinate_elements = [
        dolfinx.fem.coordinate_element(cell, 2)
        for cell in [dolfinx.mesh.CellType.quadrilateral, dolfinx.mesh.CellType.triangle]
    ]

    max_cells_per_facet = 2
    part = dolfinx.mesh.create_cell_partitioner(dolfinx.mesh.GhostMode.none, max_cells_per_facet)
    mesh = dolfinx.cpp.mesh.create_mesh(
        MPI.COMM_WORLD,
        cells_np,
        [e._cpp_object for e in coordinate_elements],
        geom,
        part,
        max_cells_per_facet,
    )
    py_mesh = Mesh(mesh, None)

    # Write mesh to file
    write_mesh("mixed_mesh_second_order.vtkhdf", py_mesh)

    # Read mesh as a 2D grid and a flat manifold in 3D
    for gdim in [2, 3]:
        mesh_in = read_mesh(MPI.COMM_WORLD, "mixed_mesh_second_order.vtkhdf", gdim=gdim)
        assert mesh_in.geometry.dim == gdim
        assert mesh_in.geometry.index_map().size_global == 12
        cmap_0 = mesh_in.geometry.cmaps[0]
        cmap_1 = mesh_in.geometry.cmaps[1]
        assert cmap_0.degree == 2
        assert cmap_1.degree == 2

        cell_types = mesh.topology.cell_types
        assert dolfinx.mesh.CellType.quadrilateral in cell_types
        assert dolfinx.mesh.CellType.triangle in cell_types


@pytest.mark.parametrize("order", [1, 2, 3])
def test_read_write_higher_order_mesh(order):
    try:
        import gmsh
    except ImportError:
        pytest.skip()

    # Create a tetrahedral mesh of a sphere
    res = 0.3
    gmsh.initialize()
    comm = MPI.COMM_WORLD
    rank = 0
    model = None
    gmsh.model.add(f"mesh_{order}")
    if comm.rank == rank:
        gmsh.option.setNumber("Mesh.CharacteristicLengthMin", res)
        gmsh.option.setNumber("Mesh.CharacteristicLengthMax", res)
        gmsh.model.occ.addSphere(0, 0, 0, 1, tag=1)
        gmsh.model.occ.synchronize()
        gmsh.model.addPhysicalGroup(3, [1], 1)
        gmsh.model.mesh.generate(3)
        gmsh.model.mesh.setOrder(order)
    comm.Barrier()

    model = comm.bcast(model, root=rank)
    # Read in mesh with gmsh to create reference dat
    ref_mesh = dolfinx.io.gmsh.model_to_mesh(gmsh.model, comm, rank).mesh
    gmsh.finalize()

    ref_volume_form = dolfinx.fem.form(
        1 * ufl.dx(domain=ref_mesh),
        dtype=ref_mesh.geometry.x.dtype,
    )
    ref_volume = comm.allreduce(dolfinx.fem.assemble_scalar(ref_volume_form), op=MPI.SUM)

    ref_surface_form = dolfinx.fem.form(
        1 * ufl.ds(domain=ref_mesh),
        dtype=ref_mesh.geometry.x.dtype,
    )
    ref_surface = comm.allreduce(dolfinx.fem.assemble_scalar(ref_surface_form), op=MPI.SUM)

    # Write to file
    filename = f"gmsh_{order}_order_sphere.vtkhdf"
    write_mesh(filename, ref_mesh)
    del ref_mesh, ref_volume_form

    # Read mesh
    mesh = read_mesh(comm, filename)

    # Compare surface and volume metrics
    volume_form = dolfinx.fem.form(1 * ufl.dx(domain=mesh), dtype=mesh.geometry.x.dtype)
    volume = comm.allreduce(dolfinx.fem.assemble_scalar(volume_form), op=MPI.SUM)
    assert np.isclose(ref_volume, volume)

    surface_form = dolfinx.fem.form(1 * ufl.ds(domain=mesh), dtype=mesh.geometry.x.dtype)
    surface = comm.allreduce(dolfinx.fem.assemble_scalar(surface_form), op=MPI.SUM)
    assert np.isclose(ref_surface, surface)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_write_point_data(dtype):
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5, dtype=dtype)
    filename = "point_data.vtkhdf"
    write_mesh(filename, mesh)
    point_data = np.arange(mesh.geometry.index_map().size_local)
    for j in range(3):
        write_point_data(filename, mesh, point_data, float(j))


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("width", [1, 3])
def test_write_cell_data(dtype, width):
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5, dtype=dtype)
    filename = "cell_data.vtkhdf"
    write_mesh(filename, mesh)
    cell_data = np.arange(mesh.topology.index_map(2).size_local * width)
    for j in range(3):
        write_cell_data(filename, mesh, cell_data, float(j))


def test_write_mixed_topology_data(mixed_topology_mesh):
    mesh = Mesh(mixed_topology_mesh, None)
    filename = "mixed_point_data.vtkhdf"
    write_mesh(filename, mesh)
    point_data = np.arange(mesh.geometry.index_map().size_local, dtype=np.float64)
    for j in range(10):
        write_point_data(filename, mesh, point_data, float(j))
        point_data *= 0.9

    filename = "mixed_cell_data.vtkhdf"
    write_mesh(filename, mesh)
    b = sum([im.size_local for im in mesh.topology.index_maps(mesh.topology.dim)])
    cell_data = np.arange(b, dtype=np.float64)
    write_cell_data(filename, mesh, cell_data, 0.0)



# Copyright (C) 2018 Michal Habera
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""General tools for timing and configuration."""

import datetime
import functools

from dolfinx import cpp as _cpp
from dolfinx.cpp.common import (
    IndexMap,
    git_commit_hash,
    hardware_concurrency,
    has_adios2,
    has_complex_ufcx_kernels,
    has_debug,
    has_kahip,
    has_parmetis,
    has_petsc,
    has_petsc4py,
    has_ptscotch,
    has_slepc,
    has_superlu_dist,
    ufcx_signature,
)

__all__ = [
    "IndexMap",
    "Reduction",
    "Timer",
    "git_commit_hash",
    "hardware_concurrency",
    "has_adios2",
    "has_complex_ufcx_kernels",
    "has_debug",
    "has_kahip",
    "has_parmetis",
    "has_petsc",
    "has_petsc4py",
    "has_ptscotch",
    "has_slepc",
    "has_superlu_dist",
    "list_timings",
    "timed",
    "timing",
    "ufcx_signature",
]

Reduction = _cpp.common.Reduction


def timing(task: str) -> tuple[int, datetime.timedelta]:
    """Return the logged elapsed time.

    Timing data is for the calling process.

    Arguments:
        task: The task name using when logging the time.

    Returns:
        (number of times logged, total wall time)
    """
    return _cpp.common.timing(task)


def list_timings(comm, reduction=Reduction.max):
    """Print out a summary of all Timer measurements.

    When used in parallel, a reduction is applied across all processes.
    By default, the maximum time is shown.
    """
    _cpp.common.list_timings(comm, reduction)


class Timer:
    """A timer for timing section of code.

    The recommended usage is with a context manager.

    Example:
        With a context manager, the timer is started when entering
        and stopped at exit. With a named :class:`Timer`::

            with Timer("Some costly operation"):
                costly_call_1()
                costly_call_2()

            delta = timing("Some costly operation")
            print(delta)

        or with an un-named :class:`Timer`::

            with Timer() as t:
                costly_call_1()
                costly_call_2()
                print(f"Elapsed time: {t.elapsed()}")

    Example:
        It is possible to start and stop a timer explicitly::

            t = Timer("Some costly operation")
            costly_call()
            delta = t.stop()

        and retrieve timing data using::

            delta = t.elapsed()

        To flush the timing data for a named :class:`Timer` to the logger,
        the timer should be stopped and flushed::

            t.stop()
            t.flush()

    Timings are stored globally (if task name is given) and once flushed
    (if used without a context manager) may be printed using functions
    :func:`timing` and :func:`list_timings`, e.g.::

        list_timings(comm)
    """

    _cpp_object: _cpp.common.Timer

    def __init__(self, name: str | None = None):
        """Create timer.

        Args:
            name: Identifier to use when storing elapsed time in logger.
        """
        self._cpp_object = _cpp.common.Timer(name)

    def __enter__(self):
        """Start timer."""
        self._cpp_object.start()
        return self

    def __exit__(self, *args):
        """Stop timer and flush timing data to logger."""
        self._cpp_object.stop()
        self._cpp_object.flush()

    def start(self) -> None:
        """Reset elapsed time and (re-)start timer."""
        self._cpp_object.start()

    def stop(self) -> datetime.timedelta:
        """Stop timer and return elapsed time.

        Returns:
            Elapsed time.
        """
        return self._cpp_object.stop()

    def resume(self) -> None:
        """Resume timer."""
        self._cpp_object.resume()

    def elapsed(self) -> datetime.timedelta:
        """Return elapsed time.

        Returns:
            Elapsed time.
        """
        return self._cpp_object.elapsed()

    def flush(self) -> None:
        """Flush timer duration to the logger.

        Note:
            Timer must have been stopped before flushing.

            Timer can be flushed only once. Subsequent calls will have
            no effect.
        """
        self._cpp_object.flush()


def timed(task: str):
    """Decorator for timing functions."""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with Timer(task):
                return func(*args, **kwargs)

        return wrapper

    return decorator



# Copyright (C) 2015-2025 Garth N. Wells, Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Unit tests for the DiscreteOperator class."""

from mpi4py import MPI

import numpy as np
import pytest

import ufl
from basix.ufl import element
from dolfinx import default_real_type
from dolfinx.cpp.mesh import cell_dim
from dolfinx.fem import Expression, Function, assemble_scalar, form, functionspace
from dolfinx.mesh import CellType, GhostMode, create_mesh, create_unit_cube, create_unit_square


@pytest.mark.petsc4py
class TestPETScDiscreteOperators:
    """Test PETSc-based discrete operators."""

    @pytest.mark.skip_in_parallel
    @pytest.mark.parametrize(
        "mesh",
        [
            create_unit_square(MPI.COMM_WORLD, 11, 6, ghost_mode=GhostMode.none),
            create_unit_square(MPI.COMM_WORLD, 11, 6, ghost_mode=GhostMode.shared_facet),
            create_unit_cube(MPI.COMM_WORLD, 4, 3, 7, ghost_mode=GhostMode.none),
            create_unit_cube(MPI.COMM_WORLD, 4, 3, 7, ghost_mode=GhostMode.shared_facet),
        ],
    )
    def test_gradient_petsc(self, mesh):
        """Test discrete gradient computation for lowest order elements."""
        from petsc4py import PETSc

        from dolfinx.fem.petsc import discrete_gradient

        V = functionspace(mesh, ("Lagrange", 1))
        W = functionspace(mesh, ("Nedelec 1st kind H(curl)", 1))
        G = discrete_gradient(V, W)
        assert G.getRefCount() == 1
        num_edges = mesh.topology.index_map(1).size_global
        m, n = G.getSize()
        assert m == num_edges
        assert n == mesh.topology.index_map(0).size_global
        G.assemble()
        assert np.isclose(G.norm(PETSc.NormType.FROBENIUS), np.sqrt(2.0 * num_edges))
        G.destroy()

    @pytest.mark.parametrize("p", range(1, 4))
    @pytest.mark.parametrize("q", range(1, 4))
    @pytest.mark.parametrize(
        "cell_type",
        [CellType.quadrilateral, CellType.triangle, CellType.tetrahedron, CellType.hexahedron],
    )
    def test_gradient_interpolation_petsc(self, cell_type, p, q):
        """Test discrete gradient computation with verification using Expression."""
        from dolfinx.fem.petsc import discrete_gradient

        comm = MPI.COMM_WORLD
        if cell_type == CellType.triangle:
            mesh = create_unit_square(
                comm, 11, 6, ghost_mode=GhostMode.none, cell_type=cell_type, dtype=default_real_type
            )
            family0 = "Lagrange"
            family1 = "Nedelec 1st kind H(curl)"
        elif cell_type == CellType.quadrilateral:
            mesh = create_unit_square(
                comm, 11, 6, ghost_mode=GhostMode.none, cell_type=cell_type, dtype=default_real_type
            )
            family0 = "Q"
            family1 = "RTCE"
        elif cell_type == CellType.hexahedron:
            mesh = create_unit_cube(
                comm,
                3,
                3,
                2,
                ghost_mode=GhostMode.none,
                cell_type=cell_type,
                dtype=default_real_type,
            )
            family0 = "Q"
            family1 = "NCE"
        elif cell_type == CellType.tetrahedron:
            mesh = create_unit_cube(
                comm,
                3,
                2,
                2,
                ghost_mode=GhostMode.none,
                cell_type=cell_type,
                dtype=default_real_type,
            )
            family0 = "Lagrange"
            family1 = "Nedelec 1st kind H(curl)"

        V = functionspace(mesh, (family0, p))
        W = functionspace(mesh, (family1, q))
        G = discrete_gradient(V, W)
        G.assemble()

        u = Function(V)
        u.interpolate(lambda x: 2 * x[0] ** p + 3 * x[1] ** p)

        grad_u = Expression(ufl.grad(u), W.element.interpolation_points)
        w_expr = Function(W)
        w_expr.interpolate(grad_u)

        # Compute global matrix vector product
        w = Function(W)
        G.mult(u.x.petsc_vec, w.x.petsc_vec)
        w.x.scatter_forward()

        atol = 100 * np.finfo(default_real_type).resolution
        assert np.allclose(w_expr.x.array, w.x.array, atol=atol)
        G.destroy()

    @pytest.mark.parametrize("p", range(1, 4))
    @pytest.mark.parametrize("q", range(1, 4))
    @pytest.mark.parametrize("from_lagrange", [True, False])
    @pytest.mark.parametrize(
        "cell_type",
        [CellType.quadrilateral, CellType.triangle, CellType.tetrahedron, CellType.hexahedron],
    )
    def test_interpolation_matrix_petsc(self, cell_type, p, q, from_lagrange):
        """Test that discrete interpolation matrix yields the same result as interpolation."""
        from dolfinx.fem.petsc import interpolation_matrix

        comm = MPI.COMM_WORLD
        if cell_type == CellType.triangle:
            mesh = create_unit_square(comm, 7, 5, ghost_mode=GhostMode.none, cell_type=cell_type)
            lagrange = "Lagrange" if from_lagrange else "DG"
            nedelec = "Nedelec 1st kind H(curl)"
        elif cell_type == CellType.quadrilateral:
            mesh = create_unit_square(comm, 11, 6, ghost_mode=GhostMode.none, cell_type=cell_type)
            lagrange = "Q" if from_lagrange else "DQ"
            nedelec = "RTCE"
        elif cell_type == CellType.hexahedron:
            mesh = create_unit_cube(comm, 3, 2, 1, ghost_mode=GhostMode.none, cell_type=cell_type)
            lagrange = "Q" if from_lagrange else "DQ"
            nedelec = "NCE"
        elif cell_type == CellType.tetrahedron:
            mesh = create_unit_cube(comm, 3, 2, 2, ghost_mode=GhostMode.none, cell_type=cell_type)
            lagrange = "Lagrange" if from_lagrange else "DG"
            nedelec = "Nedelec 1st kind H(curl)"
        v_el = element(
            lagrange, mesh.basix_cell(), p, shape=(mesh.geometry.dim,), dtype=default_real_type
        )
        s_el = element(nedelec, mesh.basix_cell(), q, dtype=default_real_type)
        if from_lagrange:
            el0 = v_el
            el1 = s_el
        else:
            el0 = s_el
            el1 = v_el

        V = functionspace(mesh, el0)
        W = functionspace(mesh, el1)
        G = interpolation_matrix(V, W)
        G.assemble()

        u = Function(V)

        def f(x):
            if mesh.geometry.dim == 2:
                return (x[1] ** p, x[0] ** p)
            else:
                return (x[0] ** p, x[2] ** p, x[1] ** p)

        u.interpolate(f)
        w_vec = Function(W)
        w_vec.interpolate(u)

        # Compute global matrix vector product
        w = Function(W)
        G.mult(u.x.petsc_vec, w.x.petsc_vec)
        w.x.scatter_forward()

        atol = 100 * np.finfo(default_real_type).resolution
        assert np.allclose(w_vec.x.array, w.x.array, atol=atol)
        G.destroy()

    @pytest.mark.skip_in_parallel
    def test_nonaffine_discrete_operator_petsc(self):
        """Test non-affine discrete operator.

        Check that discrete operator is consistent with normal
        interpolation between non-matching maps on non-affine
        geometries.
        """
        from dolfinx.fem.petsc import interpolation_matrix

        points = np.array(
            [
                [0, 0, 0],
                [1, 0, 0],
                [0, 2, 0],
                [1, 2, 0],
                [0, 0, 3],
                [1, 0, 3],
                [0, 2, 3],
                [1, 2, 3],
                [0.5, 0, 0],
                [0, 1, 0],
                [0, 0, 1.5],
                [1, 1, 0],
                [1, 0, 1.5],
                [0.5, 2, 0],
                [0, 2, 1.5],
                [1, 2, 1.5],
                [0.5, 0, 3],
                [0, 1, 3],
                [1, 1, 3],
                [0.5, 2, 3],
                [0.5, 1, 0],
                [0.5, -0.1, 1.5],
                [0, 1, 1.5],
                [1, 1, 1.5],
                [0.5, 2, 1.5],
                [0.5, 1, 3],
                [0.5, 1, 1.5],
            ],
            dtype=default_real_type,
        )

        cells = np.array([range(len(points))], dtype=np.int32)
        cell_type = CellType.hexahedron
        domain = ufl.Mesh(
            element("Lagrange", cell_type.name, 2, shape=(3,), dtype=default_real_type)
        )
        mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
        gdim = mesh.geometry.dim
        W = functionspace(mesh, ("DG", 1, (gdim,)))
        V = functionspace(mesh, ("NCE", 4))
        w, v = Function(W), Function(V)
        w.interpolate(lambda x: x)
        v.interpolate(w)

        G = interpolation_matrix(W, V)
        G.assemble()

        # Compute global matrix vector product
        v_vec = Function(V)
        G.mult(w.x.petsc_vec, v_vec.x.petsc_vec)
        v_vec.x.scatter_forward()
        atol = 10 * np.finfo(default_real_type).resolution
        assert np.allclose(v_vec.x.array, v.x.array, atol=atol)

        s = assemble_scalar(form(ufl.inner(w - v, w - v) * ufl.dx))
        assert np.isclose(s, 0, atol=atol)
        G.destroy()

    @pytest.mark.parametrize("ghost_mode", [GhostMode.none, GhostMode.shared_facet])
    @pytest.mark.parametrize(
        "cell_type",
        [CellType.triangle, CellType.quadrilateral, CellType.tetrahedron, CellType.hexahedron],
    )
    def test_discrete_interpolation(self, cell_type, ghost_mode):
        """Test discrete interpolation operator."""
        from dolfinx.fem.petsc import interpolation_matrix

        tdim = cell_dim(cell_type)
        if tdim == 2:
            mesh = create_unit_square(
                MPI.COMM_WORLD,
                4,
                4,
                cell_type=cell_type,
                ghost_mode=ghost_mode,
            )
        elif tdim == 3:
            mesh = create_unit_cube(
                MPI.COMM_WORLD,
                3,
                2,
                4,
                cell_type=cell_type,
                ghost_mode=ghost_mode,
            )
        else:
            raise ValueError(f"Unsupported {cell_type=}")

        V = functionspace(mesh, ("DG", 3))
        Q = functionspace(mesh, ("Lagrange", 3))

        u = Function(
            V,
        )
        u.interpolate(lambda x: x[0] ** 3 - x[1] ** 3)

        int_matrix = interpolation_matrix(V, Q)
        int_matrix.assemble()

        q = Function(
            Q,
        )
        int_matrix.mult(u.x.petsc_vec, q.x.petsc_vec)
        q.x.scatter_forward()

        q_ref = Function(
            Q,
        )
        q_ref.interpolate(u)

        atol = 100 * np.finfo(default_real_type).resolution
        np.testing.assert_allclose(q.x.array, q_ref.x.array, atol=atol)



# Copyright (C) 2017-2023 Garth N. Wells, Jack S. Hale
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Convert Python demos in Jupytext 'light' into MyST markdown/ipynb."""

import os
import pathlib
import runpy
import shutil

import jupytext


def process():
    """Convert Python demos in Jupytext 'light' into MyST markdown/ipynb.

    Uses Jupytext. These files can then be included in Sphinx
    documentation.
    """
    # Directories to scan
    subdirs = [pathlib.Path("../../demo")]

    # Iterate over subdirectories containing demos
    for subdir in subdirs:
        # Make demo doc directory
        demo_dir = pathlib.Path("./demos")
        demo_dir.mkdir(parents=True, exist_ok=True)

        # Process each demo using jupytext/myst
        for demo in subdir.glob("**/demo*.py"):
            python_demo = jupytext.read(demo)
            myst_text = jupytext.writes(python_demo, fmt="myst")

            # myst-parser does not process blocks with {code-cell}
            myst_text = myst_text.replace("{code-cell}", "python")
            myst_file = (demo_dir / demo.name).with_suffix(".md")
            with open(myst_file, "w") as fw:
                fw.write(myst_text)

            ipynb_file = (demo_dir / demo.name).with_suffix(".ipynb")
            jupytext.write(python_demo, ipynb_file, fmt="ipynb")

            # Copy python demo files into documentation demo directory
            shutil.copy(demo, demo_dir)

            # If demo saves matplotlib images, run the demo to create
            # images
            code = demo.read_text()
            if "savefig" in code:
                demo = demo.resolve()
                here = os.getcwd()
                os.chdir(demo_dir)
                runpy.run_path(str(demo))
                os.chdir(here)

        # Copy images used in demos
        for file in subdir.glob("**/*.png"):
            shutil.copy(file, demo_dir)


if __name__ == "__main__":
    process()



# Copyright (C) 2021-2026 Jørgen S. Dokken and Paul T. Kühner
#
# This file is part of DOLFINX (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from pathlib import Path
from xml.etree import ElementTree as ET

from mpi4py import MPI

import numpy as np
import pytest

import ufl
from basix.ufl import element
from dolfinx import default_real_type
from dolfinx.fem import Function, functionspace
from dolfinx.graph import adjacencylist
from dolfinx.mesh import CellType, create_mesh, create_unit_cube, create_unit_square


def import_adios2():
    adios2 = pytest.importorskip("adios2", minversion="2.10.0")
    if not adios2.is_built_with_mpi:
        pytest.skip("Require adios2 built with MPI support")
    return adios2


def generate_mesh(dim: int, simplex: bool, N: int = 5, dtype=None):
    """Helper function for parametrizing over meshes."""
    if dtype is None:
        dtype = default_real_type

    if dim == 2:
        if simplex:
            return create_unit_square(MPI.COMM_WORLD, N, N, dtype=dtype)
        else:
            return create_unit_square(MPI.COMM_WORLD, 2 * N, N, CellType.quadrilateral, dtype=dtype)
    elif dim == 3:
        if simplex:
            return create_unit_cube(MPI.COMM_WORLD, N, N, N, dtype=dtype)
        else:
            return create_unit_cube(MPI.COMM_WORLD, N, N, N, CellType.hexahedron, dtype=dtype)
    else:
        raise RuntimeError("Unsupported dimension")


@pytest.mark.adios2
class TestVTX:
    """Test VTXWriter functionality."""

    @pytest.mark.skipif(MPI.COMM_WORLD.size > 1, reason="This test should only be run in serial.")
    def test_second_order_vtx(self, tempdir):
        """Test saving a simple second order mesh with VTXWriter."""
        from dolfinx.io import VTXWriter

        filename = Path(tempdir, "mesh_vtx.bp")
        points = np.array([[0, 0, 0], [1, 0, 0], [0.5, 0, 0]], dtype=default_real_type)
        cells = np.array([[0, 1, 2]], dtype=np.int32)
        domain = ufl.Mesh(element("Lagrange", "interval", 2, shape=(1,), dtype=default_real_type))
        mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)
        with VTXWriter(mesh.comm, filename, mesh) as f:
            f.write(0.0)

    @pytest.mark.parametrize("dim", [2, 3])
    @pytest.mark.parametrize("simplex", [True, False])
    def test_vtx_mesh(self, tempdir, dim, simplex):
        """Test saving mesh with VTXWriter."""
        from dolfinx.io import VTXWriter

        filename = Path(tempdir, "mesh_vtx.bp")
        mesh = generate_mesh(dim, simplex)
        with VTXWriter(mesh.comm, filename, mesh) as f:
            f.write(0.0)
            mesh.geometry.x[:, 1] += 0.1
            f.write(0.1)

    @pytest.mark.parametrize("dim", [2, 3])
    @pytest.mark.parametrize("simplex", [True, False])
    def test_vtx_functions_fail(self, tempdir, dim, simplex):
        """Test for error when elements differ."""
        from dolfinx.io import VTXWriter

        mesh = generate_mesh(dim, simplex)
        gdim = mesh.geometry.dim
        v = Function(functionspace(mesh, ("Lagrange", 2, (gdim,))))
        w = Function(functionspace(mesh, ("Lagrange", 1)))
        filename = Path(tempdir, "v.bp")
        with pytest.raises(RuntimeError):
            VTXWriter(mesh.comm, filename, [v, w])

    @pytest.mark.parametrize("dim", [2, 3])
    @pytest.mark.parametrize("simplex", [True, False])
    def test_vtx_names_fail(self, tempdir, dim, simplex):
        """Test for error when elements differ."""
        from dolfinx.io import VTXWriter

        mesh = generate_mesh(dim, simplex)
        gdim = mesh.geometry.dim
        v = Function(functionspace(mesh, ("Lagrange", 1, (gdim,))), name="f")
        w = Function(functionspace(mesh, ("Lagrange", 1)), name="f")
        filename = Path(tempdir, "v.bp")
        with pytest.raises(RuntimeError):
            VTXWriter(mesh.comm, filename, [v, w])

    @pytest.mark.parametrize("simplex", [True, False])
    def test_vtx_different_meshes_function(self, tempdir, simplex):
        """Test for error when functions do not share a mesh."""
        from dolfinx.io import VTXWriter

        mesh = generate_mesh(2, simplex)
        v = Function(functionspace(mesh, ("Lagrange", 1)))
        mesh2 = generate_mesh(2, simplex)
        w = Function(functionspace(mesh2, ("Lagrange", 1)))
        filename = Path(tempdir, "v.bp")
        with pytest.raises(RuntimeError):
            VTXWriter(mesh.comm, filename, [v, w])

    @pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex64, np.complex128])
    @pytest.mark.parametrize("dim", [2, 3])
    @pytest.mark.parametrize("simplex", [True, False])
    def test_vtx_functions(self, tempdir, dtype, dim, simplex):
        """Test saving high order Lagrange functions."""
        from dolfinx.io import VTXWriter

        xtype = np.real(dtype(0)).dtype
        mesh = generate_mesh(dim, simplex, dtype=xtype)
        gdim = mesh.geometry.dim
        V = functionspace(mesh, ("DG", 2, (gdim,)))
        v = Function(V, dtype=dtype, name="v")
        bs = V.dofmap.index_map_bs

        def vel(x):
            values = np.zeros((dim, x.shape[1]), dtype=dtype)
            values[0] = x[1]
            values[1] = x[0]
            return values

        v.interpolate(vel)

        W = functionspace(mesh, ("DG", 2))
        w = Function(W, dtype=v.dtype, name="w")
        w.interpolate(lambda x: x[0] + x[1])

        filename = Path(tempdir, f"v-{np.dtype(dtype).num}.bp")
        f = VTXWriter(mesh.comm, filename, [v, w])

        # Set two cells to 0
        for c in [0, 1]:
            dofs = np.asarray([V.dofmap.cell_dofs(c) * bs + b for b in range(bs)], dtype=np.int32)
            v.x.array[dofs] = 0
            w.x.array[W.dofmap.cell_dofs(c)] = 1
        v.x.scatter_forward()
        w.x.scatter_forward()

        # Save twice and update geometry
        for t in [0.1, 1]:
            mesh.geometry.x[:, :2] += 0.1
            f.write(t)

        f.close()

    def test_save_vtkx_cell_point(self, tempdir):
        """Test writing point-wise data."""
        from dolfinx.io import VTXWriter

        mesh = create_unit_square(MPI.COMM_WORLD, 8, 5)
        P = element("Discontinuous Lagrange", mesh.basix_cell(), 0, dtype=default_real_type)

        V = functionspace(mesh, P)
        u = Function(V)
        u.interpolate(lambda x: 0.5 * x[0])
        u.name = "A"

        filename = Path(tempdir, "v.bp")
        f = VTXWriter(mesh.comm, filename, [u])
        f.write(0)
        f.close()

    def test_empty_rank_mesh(self, tempdir):
        """Test VTXWriter on mesh where some ranks have no cells."""
        from dolfinx.io import VTXWriter

        comm = MPI.COMM_WORLD
        cell_type = CellType.triangle
        domain = ufl.Mesh(
            element("Lagrange", cell_type.name, 1, shape=(2,), dtype=default_real_type)
        )

        def partitioner(comm, nparts, local_graph, num_ghost_nodes):
            """Leave cells on the current rank."""
            dest = np.full(len(cells), comm.rank, dtype=np.int32)
            return adjacencylist(dest)._cpp_object

        if comm.rank == 0:
            cells = np.array([[0, 1, 2], [0, 2, 3]], dtype=np.int64)
            x = np.array([[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [0.0, 1.0]], dtype=default_real_type)
        else:
            cells = np.empty((0, 3), dtype=np.int64)
            x = np.empty((0, 2), dtype=default_real_type)

        mesh = create_mesh(comm, cells, domain, x, partitioner)

        V = functionspace(mesh, ("Lagrange", 1))
        u = Function(V)

        filename = Path(tempdir, "empty_rank_mesh.bp")
        with VTXWriter(comm, filename, u) as f:
            f.write(0.0)

    @pytest.mark.parametrize("dim", [2, 3])
    @pytest.mark.parametrize("simplex", [True, False])
    @pytest.mark.parametrize("reuse", [True, False])
    def test_vtx_reuse_mesh(self, tempdir, dim, simplex, reuse):
        """Test reusage of mesh by VTXWriter."""
        from dolfinx.io import VTXMeshPolicy, VTXWriter

        adios2 = import_adios2()

        mesh = generate_mesh(dim, simplex)
        v = Function(functionspace(mesh, ("Lagrange", 1)))
        filename = Path(tempdir, "v.bp")
        v.name = "v"
        policy = VTXMeshPolicy.reuse if reuse else VTXMeshPolicy.update

        # Save three steps
        writer = VTXWriter(mesh.comm, filename, v, "BP4", policy)
        writer.write(0)
        v.interpolate(lambda x: 0.5 * x[0])
        writer.write(1)
        v.interpolate(lambda x: x[1])
        writer.write(2)
        writer.close()

        reuse_variables = ["NumberOfEntities", "NumberOfNodes", "connectivity", "geometry", "types"]
        target_all = 3  # For all other variables the step count is number of writes
        target_mesh = 1 if reuse else 3
        # For mesh variables the step count is 1 if reuse else number of writes
        adios = adios2.Adios(comm=mesh.comm)
        io = adios.declare_io("TestData")
        io.set_engine("BP4")
        adios_file = adios2.Stream(io, str(filename), "r", mesh.comm)

        for name, var in adios_file.available_variables().items():
            if name in reuse_variables:
                assert int(var["AvailableStepsCount"]) == target_mesh
            else:
                assert int(var["AvailableStepsCount"]) == target_all
        adios_file.close()

    def test_dg_0_data(self, tempdir):
        """Test that we can mix DG-0 and other Lagrange functions."""
        from dolfinx.io import VTXWriter

        adios2 = import_adios2()
        mesh = generate_mesh(2, False)
        v = Function(functionspace(mesh, ("Lagrange", 2, (2,))))
        filename = Path(tempdir, "v.bp")
        v.name = "v"
        v.interpolate(lambda x: (x[0], -x[1]))

        q = Function(functionspace(mesh, ("DG", 0, (2,))))
        q.name = "q"
        q.x.array[:] = np.arange(q.x.array.size, dtype=q.x.array.dtype)

        # Save three steps
        writer = VTXWriter(mesh.comm, filename, [v, q])
        writer.write(0)
        v.interpolate(lambda x: (0.5 * x[0], x[1]))
        writer.write(1)
        q.interpolate(lambda x: (x[1], x[0]))
        writer.write(2)
        writer.close()

        # Check that q and v were written as point and cell data respectively
        with adios2.FileReader(str(filename), mesh.comm) as bpFile:
            vtk_data = bpFile.read_attribute_string("vtk.xml")

            vtk_data = ET.fromstring(vtk_data).find("UnstructuredGrid/Piece")
            assert vtk_data is not None

            xml_point_data = vtk_data.find("PointData")
            xml_cell_data = vtk_data.find("CellData")
            assert xml_point_data is not None
            assert xml_cell_data is not None

            def get_data_array_names(data):
                return set(da.get("Name") for da in data.findall("DataArray"))

            point_arrays = get_data_array_names(xml_point_data)
            assert "v" in point_arrays or ("v_imag" in point_arrays and "v_real" in point_arrays)

            cell_arrays = get_data_array_names(xml_cell_data)
            assert "q" in cell_arrays or ("q_imag" in cell_arrays and "q_real" in cell_arrays)
